﻿#include "CMQObject.h"
#include <algorithm>
#include <unordered_map>

namespace cmq
{

	// 2つの頂点インデックスから辺IDを作る
	lindex_t makeLineId(vindex_t v1, vindex_t v2)
	{
		return (v1 < v2) ? ((static_cast<int64_t>(v1) << 32) + v2) : ((static_cast<int64_t>(v2) << 32) + v1);
	}

	// 頂点インデックスのpairから辺IDを作る
	lindex_t makeLineId(std::pair<vindex_t, vindex_t> line)
	{
		return makeLineId(line.first, line.second);
	}

	// 辺IDから頂点インデックスのpairを作る
	std::pair<vindex_t, vindex_t> lineId2VertexPair(lindex_t lindex)
	{
		return std::make_pair<vindex_t, vindex_t>(lindex >> 32, lindex & 0xffff);
	}

	// 面インデックスと辺インデックスから面辺IDを作る
	flindex_t makeFaceLineId(findex_t f, int l)
	{
		return ((static_cast<int64_t>(f) << 32) + l);
	}

	// 面辺IDから面インデックスと辺インデックスのpairを作る
	std::pair<findex_t, int> flId2FaceLinePair(flindex_t flindex)
	{
		return lineId2VertexPair(flindex);
	}

	const float CMQCoordinate::uvdelta = 0.0001f;

	CMQObject &CMQDocument::insertObject(CMQObject &newobj, const CMQObject &before)
	{
		int id = m_doc->InsertObject(newobj.m_obj, before.m_obj);
		newobj.m_obj->SetDepth(before.m_obj->GetDepth());
		newobj.m_obj->SetVisible(before.m_obj->GetVisible());
		newobj.setDocument(*this);
		return newobj;
	}

	CMQObject CMQDocument::getCurrentObject()
	{
		int idx = m_doc->GetCurrentObjectIndex();
		return CMQObject(m_doc->GetObject(idx), *this);
	}
	void CMQDocument::deleteObject(CMQObject &obj)
	{
		oindex_t i = getObjectIndex(obj);
		if(i == -1)
			obj.m_obj->DeleteThis();
		else
			m_doc->DeleteObject(i);
	}

	CMQObject CMQDocument::getObject(oindex_t i) { return CMQObject(m_doc->GetObject(i), *this); }
	CMQObject CMQDocument::getObjectFromUniqueID(UINT uid) const { return CMQObject(m_doc->GetObjectFromUniqueID(uid), *this); }
	oindex_t CMQDocument::getObjectIndex(CMQObject &obj) const
	{
		if(!obj.m_obj) {
			return -1;
		}
		return m_doc->GetObjectIndex(obj.m_obj);
	}

	bool CMQDocument::isSelectVertex(CMQObject &obj, vindex_t vindex) const { return !!m_doc->IsSelectVertex(m_doc->GetObjectIndex(obj.m_obj), vindex); }
	bool CMQDocument::addSelectVertex(CMQObject &obj, vindex_t vindex) { return !!m_doc->AddSelectVertex(m_doc->GetObjectIndex(obj.m_obj), vindex); }
	bool CMQDocument::deleteSelectVertex(CMQObject &obj, vindex_t vindex) { return !!m_doc->DeleteSelectVertex(m_doc->GetObjectIndex(obj.m_obj), vindex); }

	bool CMQDocument::isSelectFace(CMQObject &obj, findex_t findex) const { return !!m_doc->IsSelectFace(m_doc->GetObjectIndex(obj.m_obj), findex); }
	bool CMQDocument::addSelectFace(CMQObject &obj, findex_t findex) { return !!m_doc->AddSelectFace(m_doc->GetObjectIndex(obj.m_obj), findex); }
	bool CMQDocument::deleteSelectFace(CMQObject &obj, findex_t findex) { return !!m_doc->DeleteSelectFace(m_doc->GetObjectIndex(obj.m_obj), findex); }

	bool CMQDocument::isSelectLine(CMQObject &obj, findex_t findex, int lindex) const { return !!m_doc->IsSelectLine(m_doc->GetObjectIndex(obj.m_obj), findex, lindex); }
	bool CMQDocument::addSelectLine(CMQObject &obj, findex_t findex, int lindex) { return !!m_doc->AddSelectLine(m_doc->GetObjectIndex(obj.m_obj), findex, lindex); }
	bool CMQDocument::deleteSelectLine(CMQObject &obj, findex_t findex, int lindex) { return !!m_doc->DeleteSelectLine(m_doc->GetObjectIndex(obj.m_obj), findex, lindex); }

	bool CMQDocument::isSelectUVVertex(CMQObject &obj, findex_t findex, int vindex) const { return !!m_doc->IsSelectUVVertex(m_doc->GetObjectIndex(obj.m_obj), findex, vindex); }
	bool CMQDocument::addSelectUVVertex(CMQObject &obj, findex_t findex, int vindex) { return !!m_doc->AddSelectUVVertex(m_doc->GetObjectIndex(obj.m_obj), findex, vindex); }
	bool CMQDocument::deleteSelectUVVertex(CMQObject &obj, findex_t findex, int vindex) { return !!m_doc->DeleteSelectUVVertex(m_doc->GetObjectIndex(obj.m_obj), findex, vindex); }

	bool CMQObject::isSelectVertex(int index) { return m_doc.isSelectVertex(*this, index); }
	bool CMQObject::addSelectVertex(int index) { return m_doc.addSelectVertex(*this, index); }
	bool CMQObject::deleteSelectVertex(int index) { return m_doc.deleteSelectVertex(*this, index); }
	bool CMQObject::isSelectFace(int index) { return m_doc.isSelectFace(*this, index); }
	bool CMQObject::addSelectFace(int index) { return m_doc.addSelectFace(*this, index); }
	bool CMQObject::deleteSelectFace(int index) { return m_doc.deleteSelectFace(*this, index); }
	bool CMQObject::isSelectUVVertex(findex_t findex, int index) { return m_doc.isSelectUVVertex(*this, findex, index); }
	bool CMQObject::addSelectUVVertex(findex_t findex, int index) { return m_doc.addSelectUVVertex(*this, findex, index); }
	bool CMQObject::deleteSelectUVVertex(findex_t findex, int index) { return m_doc.deleteSelectUVVertex(*this, findex, index); }

	bool CMQObject::isLive() const
	{
#if _DEBUG
		if(!m_obj) {
			OutputDebugString(L"invalid m_obj\n");
		}
		if(!m_doc.m_doc) {
			OutputDebugString(L"invalid m_doc\n");
		}
#endif
		return !!m_obj && !!m_doc.m_doc && m_doc.getObjectIndex(*const_cast<CMQObject*>(this)) >= 0;
	}

	CMQObject CMQObject::clone()
	{
		return CMQObject(m_obj->Clone(), m_doc);
	}

	/*
		MQObjectの名前を取得してstringで返す
		*/
	std::string CMQObject::getName()
	{
		const int NAME_BUFFER_SIZE = 250;
		char objname[NAME_BUFFER_SIZE];
		m_obj->GetName(objname, NAME_BUFFER_SIZE);
		return std::string(objname);
	}

	void CMQObject::setName(const std::string &name)
	{
		m_obj->SetName(name.c_str());
	}

	/*
		nameをベースとして使われていない名前を取得してstringで返す
		*/
	std::string CMQDocument::getUnusedName(const std::string &name)
	{
		const int NAME_BUFFER_SIZE = 250;
		char unusedname[NAME_BUFFER_SIZE];
		m_doc->GetUnusedObjectName(unusedname, NAME_BUFFER_SIZE, name.c_str());
		return std::string(unusedname);
	}

	CMQObject CMQDocument::getObjectbyName(const std::string &name)
	{
		for(int i = 0; i < m_doc->GetObjectCount(); i++) {
			CMQObject obj = getObject(i);
			if(obj.isLive() && obj.getName() == name) {
				return obj;
			}
		}
		return CMQObject();
	}

	std::string CMQDocument::getMaterialName(MQMaterial mat) const
	{
		const int NAME_BUFFER_SIZE = 250;
		char name[NAME_BUFFER_SIZE];
		mat->GetName(name, NAME_BUFFER_SIZE);
		return std::string(name);
	}

	std::string CMQDocument::getTextureName(MQMaterial mat) const
	{
		const int NAME_BUFFER_SIZE = 250;
		char name[NAME_BUFFER_SIZE];
		mat->GetTextureName(name, NAME_BUFFER_SIZE);
		return std::string(name);
	}

	// 頂点を含む辺のリスト
	std::vector<int> CMQObject::getVertexRelatedFaces(int vindex, std::vector<int> &faces) const
	{
		int fcount = m_obj->GetVertexRelatedFaces(vindex, nullptr);
		if(fcount == 0) {
			faces.clear();
			return faces;
		}
		faces.resize(fcount);
		m_obj->GetVertexRelatedFaces(vindex, &faces[0]);
		return faces;
	}

	// 面を構成する頂点のリスト
	std::vector<int> CMQObject::getFacePointArray(int findex, std::vector<int> &vertexes) const
	{
		int vcount = m_obj->GetFacePointCount(findex);
		if(vcount == 0) {
			vertexes.clear();
			return vertexes;
		}
		vertexes.resize(vcount);
		m_obj->GetFacePointArray(findex, &vertexes[0]);
		return vertexes;
	}

	// 面を構成する頂点のUV値リスト
	std::vector<CMQCoordinate> CMQObject::getFaceCoordinateArray(int findex, std::vector<CMQCoordinate> &coord) const
	{
		int ccount = m_obj->GetFacePointCount(findex);
		if(ccount == 0) {
			coord.clear();
			return coord;
		}
		std::unique_ptr<int[]> vertexarray(new int[ccount]);
		m_obj->GetFacePointArray(findex, vertexarray.get());
		std::unique_ptr<MQCoordinate[]> coodarry(new MQCoordinate[ccount]);
		m_obj->GetFaceCoordinateArray(findex, coodarry.get());
		coord.clear();
		coord.reserve(ccount);
		for(int i = 0; i < ccount; i++) {
			coord.push_back(CMQCoordinate(vertexarray[i], coodarry[i]));
		}
		return coord;
	}

	// 面を構成する指定頂点のUVセット
	void CMQObject::setFaceCoordinate(int findex, int vertex, const CMQCoordinate &co)
	{
		int vcount = m_obj->GetFacePointCount(findex);
		std::unique_ptr<MQCoordinate[]> uvarray(new MQCoordinate[vcount]);
		m_obj->GetFaceCoordinateArray(findex, uvarray.get());
		uvarray[vertex] = co.getCoordinate();
		m_obj->SetFaceCoordinateArray(findex, uvarray.get());
	}

	// 面を構成する頂点のUVを一括でセット
	void CMQObject::setFaceCoordinateArray(int findex, const std::vector<CMQCoordinate> &carray)
	{
		std::unique_ptr<MQCoordinate[]> pcarray(new MQCoordinate[carray.size()]);
		for(std::vector<CMQCoordinate>::size_type i = 0; i < carray.size(); i++) {
			pcarray[i] = carray[i].getCoordinate();
		}
		m_obj->SetFaceCoordinateArray(findex, pcarray.get());
	}

	// 面を作る
	int CMQObject::addFace(std::vector<int> varray)
	{
		return m_obj->AddFace(varray.size(), &varray[0]);
	}

	// 面インデックスと辺インデックスで指定したラインの両端点の頂点インデックスを取得
	std::pair<vindex_t, vindex_t> CMQObject::getLineVertex(findex_t findex, int lindex) const
	{
		auto varray = getFacePointArray(findex);
		if(lindex < 0) { lindex += varray.size(); }
		lindex %= varray.size();
		int v1 = varray[lindex];
		int v2 = varray[lindex == varray.size() - 1 ? 0 : lindex + 1];
		return std::make_pair(v1, v2);
	}

	// 面インデックスと辺インデックスで指定したラインをv0を始点としたベクトルで取得
	MQPoint CMQObject::getLineVector(findex_t findex, int lindex, vindex_t v0) const
	{
		auto vpair = getLineVertex(findex, lindex);
		MQPoint v1 = getVertex(vpair.first);
		MQPoint v2 = getVertex(vpair.second);
		MQPoint vec = (v0 == -1 || v0 == vpair.first) ? v2 - v1 : v1 - v2;
		return vec;
	}

	// 面の指定した辺における法線ベクトルを取得
	MQPoint CMQObject::getNormalVector(int findex, int lindex) const
	{
		MQPoint vec1 = getLineVector(findex, lindex);
		MQPoint vec2 = getLineVector(findex, lindex - 1);
		if(getFacePointCount(findex) > 3) {
			MQPoint vec3 = getLineVector(findex, lindex + 1);
			vec2 -= vec3;
			vec2 /= 2;
		}
		MQPoint normal = CMQPoint::outercross(vec1, vec2);
		return normal;
	}

	// 隣接頂点
	MQPoint CMQObject::getNeiborPoint(int findex, int lindex) const
	{
		MQPoint vec2 = -getLineVector(findex, lindex - 1);
		if(getFacePointCount(findex) > 3) {
			MQPoint vec3 = getLineVector(findex, lindex + 1);
			vec2 += vec3;
			vec2 /= 2;
		}
		int vindex = getFacePointArray(findex)[lindex];
		MQPoint v1 = getVertex(vindex);
		return v1 + vec2;
	}

	// 指定頂点の隣接辺
	std::unordered_set<flindex_t> CMQObject::getNeighborLine(vindex_t vindex, lindex_t lindex) const
	{
		auto vpair = lineId2VertexPair(lindex);
		vindex_t previndex = (vpair.first == vindex) ? vpair.second : vpair.first;
		std::unordered_set<flindex_t> lineset;
		auto farray = getVertexRelatedFaces(vindex);
		std::vector<int> vertexes(5);
		for(findex_t findex : farray) {
			auto varray = getFacePointArray(findex, vertexes);
			if(varray.size() == 2) {
				if(makeLineId(varray[0], varray[1]) != lindex) {
					lineset.insert(makeFaceLineId(findex, 0));
				}
			}
			else {
				for(std::vector<int>::size_type i = 0; i < varray.size(); i++) {
					if(varray[i] == vindex) {
						if(varray[(i + 1) % varray.size()] != previndex) {
							lineset.insert(makeFaceLineId(findex, i));
						}
						if(varray[(i - 1 + varray.size()) % varray.size()] != previndex) {
							lineset.insert(makeFaceLineId(findex, i == 0 ? varray.size() - 1 : i - 1));
						}
						break;
					}
				}
			}
		}
		return lineset;
	}

	// 2座標を結ぶラインを作成
	int CMQObject::createLine(MQPoint &p1, MQPoint &p2)
	{
		int v1 = addVertex(p1);
		int v2 = addVertex(p2);
		std::vector<int> varray;
		varray.push_back(v1);
		varray.push_back(v2);
		int lindex = addFace(varray);
		return lindex;
	}

	// 指定の面の指定辺以外を取得
	std::vector<std::pair<int, int>> CMQObject::getOtherLine(std::pair<vindex_t, vindex_t> line, int face, bool selectonly)
	{
		return getOtherLine(line.first, line.second, face, selectonly);
	}

	// 指定の面の指定辺以外を取得
	std::vector<std::pair<int, int>> CMQObject::getOtherLine(int v1, int v2, int face, bool selectonly)
	{
		std::vector<std::pair<int, int>> lines;
		int fcount = getFaceCount();
		std::vector<int> vertexes(5);
		for(int findex = 0; findex < fcount; findex++) {
			if(face == findex) {
				continue;
			}
			if(selectonly && !isSelectFace(findex)) {
				continue;
			}
			auto varray = getFacePointArray(findex, vertexes);
			int vcount = varray.size();
			int vsize = vcount == 2 ? 1 : vcount;
			for(int lindex = 0; lindex < vsize; lindex++) {
				int v3 = varray[lindex];
				int v4 = varray[(lindex + 1) % vcount];
				if((v1 == v3 && v2 == v4) || (v1 == v4 && v2 == v3)) {
					lines.push_back(std::make_pair(findex, lindex));
				}
			}
		}
		return lines;
	}

	// 指定辺の指定頂点を含む辺のリスト
	std::vector<lindex_t> CMQObject::getLineArray(findex_t findex, vindex_t vindex)
	{
		auto varray = getFacePointArray(findex);
		std::vector<lindex_t> larray(varray.size());
		vindex_t v1 = varray.back();
		for(vindex_t v : varray) {
			if(vindex == -1 || v1 == vindex || v == vindex) {
				larray.push_back(makeLineId(v1, v));
			}
			v1 = v;
		}
		return larray;
	}

	// 面を移動（指定座標の面に作り直す）
	int CMQObject::modifyFace(int index, const std::vector<int> &varray)
	{
		if(varray.size() != getFacePointCount(index)) {
			deleteFace(index);
			return -1;
		}
		int newindex = addFace(varray);
		setFaceVisible(newindex, getFaceVisible(index));
		setFaceMaterial(newindex, getFaceMaterial(index));
		auto carray = getFaceCoordinateArray(index);
		setFaceCoordinateArray(newindex, carray);
		for(std::vector<int>::size_type i = 0; i < varray.size(); i++) {
			setFaceVertexColor(newindex, i, getFaceVertexColor(index, i));
			setFaceEdgeCrease(newindex, i, getFaceEdgeCrease(index, i));
		}
		deleteFace(index);
		return newindex;
	}

	// 面の属性をコピー
	void CMQObject::copyFaceAttribute(findex_t findex, const CMQObject &object, findex_t findex2, std::vector<cmq::CMQCoordinate> &coord)
	{
		// material
		setFaceMaterial(findex, object.getFaceMaterial(findex2));
		// visible
		setFaceVisible(findex, object.getFaceVisible(findex2));
		// coodinate
		setFaceCoordinateArray(findex, object.getFaceCoordinateArray(findex2, coord));
		//vertex color

		// edge crease

	}

	// オブジェクトをコピー
	void CMQObject::copyFromObject(const CMQObject &object, const MQMatrix *r)
	{
		if(object.m_obj == m_obj) {
			return;
		}
		// objectの頂点をコピーする
		std::vector<vindex_t> vmap(object.getVertexCount(), -1);
		cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(const_cast<cmq::CMQObject*>(&object));
		for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
			if(vit->getVertexRefCount() == 0) {
				continue;
			}
			MQPoint p = vit->p();
			if(r) {
				p = p * (*r);
			}
			cmq::vindex_t vindex = addVertex(p);
			copyVertexAttribute(vindex, object, vit.index());
			vmap[vit.index()] = vindex;
		}

		std::vector<int> vertexes(5);
		std::vector<cmq::vindex_t> newvarray(5);
		std::vector<cmq::CMQCoordinate> coord(5);
		// objectの面をコピーする
		cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(const_cast<cmq::CMQObject*>(&object));
		for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
			auto varray = fit->getFacePointArray(vertexes);
			if(varray.empty()) {
				continue;
			}
			newvarray.resize(varray.size());
			std::transform(varray.begin(), varray.end(), newvarray.begin(), [&vmap](cmq::vindex_t vindex) {
				return (vindex < 0 || (std::vector<cmq::vindex_t>::size_type)vindex >= vmap.size()) ? -1 : vmap[vindex];
			});
			findex_t findex = addFace(newvarray);
			copyFaceAttribute(findex, object, fit.index(), coord);
		}
	}

	//-------------------------------------------------------------------------------------
	// CMQObject
	//-------------------------------------------------------------------------------------
	// <class CMQVertex>
	template <> int CMQObject::getCount<CMQVertex>() const { return getVertexCount(); }

	template <> bool CMQObject::isValid<CMQVertex>(int index) const { return getVertexRefCount(index) > 0; }
	template <> int CMQObject::getCount<CMQFace>() const { return getFaceCount(); }

	template <> bool CMQObject::isValid<CMQFace>(int index) const { return getFacePointCount(index) > 0; }

	template <> int CMQDocument::getCount<CMQObject>() const { return getObjectCount(); }
	/*
	template <> bool CMQDocument::isValid<CMQObject>(int index) const
	{
	//	return const_cast<CMQDocument>operator[]->getObject(index);
	//	return const_cast<CMQDocument*>(this)->getObject(index);
	return !!m_doc->GetObject(index);
	}
	*/
	int CMQFace::getFacePointCount() const { return m_obj->getFacePointCount(m_index); }

	std::vector<int> CMQFace::getFacePointArray(std::vector<int> &vertexes) const { return m_obj->getFacePointArray(m_index, vertexes); }

	std::vector<int> CMQVertex::getVertexRelatedFaces() const { return m_obj->getVertexRelatedFaces(m_index); }

	int CMQVertex::getVertexRefCount() const { return m_obj->getVertexRefCount(m_index); }

	bool CMQVertex::isSelect() const { return m_obj->isSelectVertex(m_index); }
	void CMQVertex::addSelect() const { m_obj->addSelectVertex(m_index); }
	void CMQVertex::deleteSelect() const { m_obj->deleteSelectVertex(m_index); }
	bool CMQFace::isSelect() const { return m_obj->isSelectFace(m_index); }
	void CMQFace::addSelect() const { m_obj->addSelectFace(m_index); }
	void CMQFace::deleteSelect() const { m_obj->deleteSelectFace(m_index); }

	void CMQFace::copyFaceAttribute(const CMQObject &object, findex_t findex2) { m_obj->copyFaceAttribute(m_index, object, findex2); }

	std::vector<TrianglePoints> CMQObject::triangulate(findex_t findex)
	{
		auto varray = getFacePointArray(findex);
		std::vector<MQPoint> parray;
		parray.reserve(varray.size());
		for(vindex_t vindex : varray) {
			parray.push_back(getVertex(vindex));
		}
		return m_doc.triangulate(parray);
	}

	std::vector<TrianglePoints> CMQDocument::triangulate(std::vector<MQPoint> &parray)
	{
		std::vector<int> array((parray.size() - 2) * 3);
		BOOL result = m_doc->Triangulate(&parray[0], parray.size(), &array[0], array.size());
		std::vector<TrianglePoints> resultarray;
		if(result) {
			resultarray.resize(parray.size() - 2);
			for(int i = 0; i < parray.size() - 2; i++) {
				for(int j = 0; j < 3; j++) {
					resultarray[i].idx[j] = array[i * 3 + j];
				}
			}
		}
		return resultarray;
	}

	//CMQIterator
	/*
	template<>
	const CMQVertex &CMQIterator<CMQVertex>::get() const {
	CMQVertex v(m_container->getVertex(m_index), m_container, m_index);
	return std::move(v);
	}

	template<>
	const CMQFace &CMQIterator<CMQFace>::get() const {
	std::shared_ptr<CMQFace> p(new CMQFace(m_container, m_index));
	return *p;
	}
	*/
	template<>
	std::shared_ptr<CMQVertex> CMQIterator<CMQVertex>::getp() const
	{
		std::shared_ptr<CMQVertex> p(new CMQVertex(m_container->getVertex(m_index), m_container, m_index));
		return p;
	}

	template<>
	std::shared_ptr<CMQFace> CMQIterator<CMQFace>::getp() const
	{
		std::shared_ptr<CMQFace> p(new CMQFace(m_container, m_index));
		return p;
	}

	//---------------------------------------------------------------------------------
	std::wstring towstring(const std::string &s)
	{
		int buflen = s.length() * 3;
		std::unique_ptr<wchar_t[]> wcs(new wchar_t[buflen + 1]);
		size_t len = 0;
		errno_t err = mbstowcs_s(&len, wcs.get(), buflen, s.c_str(), _TRUNCATE);
		if(err) {
			return L"invalid";
		}
		return std::wstring(wcs.get());
	}

	std::string tostring(const std::wstring &s)
	{
		int buflen = s.length() * 2;
		std::unique_ptr<char[]> mbs(new char[buflen + 1]);
		size_t len = 0;
		errno_t err = wcstombs_s(&len, mbs.get(), buflen, s.c_str(), _TRUNCATE);
		if(err) {
			return "invalid";
		}
		return std::string(mbs.get());
	}
}
