//---------------------------------------------------------------------------
//
//   UVToolsX
//
//          Copyright(C) 2014, Tark
//
//  UVToolsX for Metasequoia V4
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <memory>
#include <vector>
#include <sstream>
#include <algorithm>

#include "UVToolsXWindow.h"
#include "UVToolsX.h"
#include <MQ3DLib.h>
#include <MQSetting.h>

static UVToolsX plugin;

// Constructor
// RXgN^
UVToolsX::UVToolsX() : m_Window(nullptr), m_bActivate(false), m_texturemode(TEX_WIRE | TEX_CUROBJ | TEX_CURMAT), m_texturewidth(-1), m_textureheight(-1)
{
	setlocale(LC_CTYPE, "");
}

// Destructor
// fXgN^
UVToolsX::~UVToolsX()
{
	if(m_Window != NULL) {
		delete m_Window;
		m_Window = NULL;
	}
}

//---------------------------------------------------------------------------
//  GetPlugInID
//---------------------------------------------------------------------------
void UVToolsX::GetPlugInID(DWORD *Product, DWORD *ID)
{
	*Product = 0xE3F7A830;
	*ID = 0xBD9A1875;
}

//---------------------------------------------------------------------------
//  GetPlugInName
//---------------------------------------------------------------------------
const char *UVToolsX::GetPlugInName(void)
{
	return "UVToolsX for V4 v.1.2      Copyright(C) 2014, Tark.";
}

//---------------------------------------------------------------------------
//  EnumString
//---------------------------------------------------------------------------
const char *UVToolsX::EnumString(void)
{
	return "UVToolsX";
}

//---------------------------------------------------------------------------
//  Initialize
//---------------------------------------------------------------------------
BOOL UVToolsX::Initialize()
{
	int posX = -1, posY = -1;
	int texturewidth = -1, textureheight = -1;
	{
		MQSetting *setting = OpenSetting();
		setting->Load("PosX", posX, -1);
		setting->Load("PosY", posY, -1);
		setting->Load("TexWidth", texturewidth, 1024);
		setting->Load("TexHeight", textureheight, 1024);
		CloseSetting(setting);
	}
	if(m_Window == NULL) {
		MQWindowBase mainwnd = MQWindow::GetMainWindow();
		m_Window = new UVToolsXWindow(this, mainwnd);
		if(posX >= 0 && posY >= 0) {
			m_Window->SetPosX(posX);
			m_Window->SetPosY(posY);
		}
		m_texturewidth = texturewidth;
		m_textureheight = textureheight;
	}
	Activate(nullptr, m_bActivate);
	return TRUE;
}

//---------------------------------------------------------------------------
//  Exit
//    AvP[V̏I
//---------------------------------------------------------------------------
void UVToolsX::Exit()
{
	{
		MQSetting *setting = OpenSetting();
		setting->Save("PosX", m_Window->GetPosX());
		setting->Save("PosY", m_Window->GetPosY());
		setting->Save("TexWidth", m_texturewidth);
		setting->Save("TexHeight",m_textureheight);
		CloseSetting(setting);
	}
	if(m_Window != NULL) {
		delete m_Window;
		m_Window = NULL;
	}
}

//---------------------------------------------------------------------------
//  Activate
//---------------------------------------------------------------------------
BOOL UVToolsX::Activate(MQDocument doc, BOOL flag)
{
	//	m_doc = doc;
	if(m_Window == NULL) {
		return FALSE;
	}
	m_bActivate = flag ? true : false;
	m_Window->SetVisible(m_bActivate);
	m_Window->resize();

	return m_bActivate;
}

//---------------------------------------------------------------------------
//  IsActivated
//---------------------------------------------------------------------------
BOOL UVToolsX::IsActivated(MQDocument doc)
{
	if(m_Window == NULL) {
		return FALSE;
	}
	return m_bActivate;
}

//---------------------------------------------------------------------------
//  ExecuteCallback
///---------------------------------------------------------------------------
bool UVToolsX::ExecuteCallback(MQDocument doc, void *option)
{
	CallbackInfo *info = (CallbackInfo*)option;
	return ((*this).*info->proc)(doc);
}

// R[obŇĂяo
void UVToolsX::Execute(ExecuteCallbackProc proc)
{
	CallbackInfo info;
	info.proc = proc;
	BeginCallback(&info);
}

// vOC̃x[XNXԂ
MQBasePlugin *GetPluginClass()
{
	return &plugin;
}

// ʂ̍\_t[eBOI
BOOL UVToolsX::selectFloatingUV(MQDocument doc_)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		// SUV_IĂʈȊOUV_̑I
		auto vcount = fit->getFacePointCount();
		bool selected_all = true;
		for(int i = 0; i < vcount; i++) {
			if(!obj.isSelectUVVertex(fit.index(), i)) {
				selected_all = false;
				break;
			}
		}
		if(!selected_all) {
			for(int i = 0; i < vcount; i++) {
				obj.deleteSelectUVVertex(fit.index(), i);
			}
		}
	}
	return TRUE;
}

// _UV_I(IȂƂz)
BOOL UVToolsX::selectUVsame(MQDocument doc_)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto varray = fit->getFacePointArray();
		for(std::vector<cmq::vindex_t>::size_type i = 0; i < varray.size(); i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				// IUV_̒_
				cmq::vindex_t vindex = varray[i];
				std::vector<cmq::flindex_t> uvlist = getSameUV(obj, vindex);
				for(cmq::flindex_t flindex : uvlist) {
					obj.addSelectUVVertex(flindex);
				}
			}
		}
	}
	return TRUE;
}

// _UV_Xg擾
std::vector<cmq::flindex_t> UVToolsX::getSameUV(cmq::CMQObject &obj, cmq::vindex_t vindex)
{
	CursorKeeper cur;
	std::vector<cmq::flindex_t> uvlist;
	std::vector<int> vertexbuf(5);
	auto farray = obj.getVertexRelatedFaces(vindex);
	// eʂ̊YUV_I
	for(auto findex : farray) {
		auto varray = obj.getFacePointArray(findex, vertexbuf);
		for(std::vector<cmq::vindex_t>::size_type j = 0; j < varray.size(); j++) {
			if(varray[j] == vindex) {
				uvlist.push_back(cmq::makeFaceLineId(findex, j));
				break;
			}
		}
	}
	return uvlist;
}

// O[vUV_I
BOOL UVToolsX::selectGroupUV(MQDocument doc_)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	// T
	std::unordered_set<cmq::findex_t> foundUV;
	// UV_O[v
	searchUVisland(obj, foundUV);
	// TʂUV_I
	for(cmq::findex_t findex : foundUV) {
		int vcount = obj.getFacePointCount(findex);
		for(int i = 0; i < vcount; i++) {
			obj.addSelectUVVertex(findex, i); // UV_I
		}
	}
	return TRUE;
}

// _UV}bv쐬
std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> UVToolsX::makev2uvmap(cmq::CMQObject &obj)
{
	std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> v2uvmap;
	v2uvmap.reserve(obj.getFaceCount() * 5);
	std::vector<int> vertexbuf(5);
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto varray = fit->getFacePointArray(vertexbuf);
		for(std::vector<cmq::vindex_t>::size_type i = 0; i < varray.size(); i++) {
			cmq::flindex_t flindex = cmq::makeFaceLineId(fit.index(), i);
			v2uvmap.insert(std::make_pair(varray[i], flindex));
		}
	}
	return v2uvmap;
}

// O[vUV_
BOOL UVToolsX::searchUVisland(cmq::CMQObject &obj, std::unordered_set<cmq::findex_t> &foundUV)
{
	// _UV}bv
	std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> v2uvmap(makev2uvmap(obj));
	// ςݖ
	std::unordered_set<cmq::findex_t> selectedface;
	selectedface.reserve(obj.getFaceCount());
	std::vector<int> vertexbuf(5);
	//	std::vector<CMQCoordinate> coordbuf(5);
	// TJnʂ邽߂ɑIUV_
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		// ʂTς݂ȂXLbv
		if(selectedface.find(fit.index()) != selectedface.end()) {
			continue;
		}
		// ʂTςmapɑ}
		auto varray = fit->getFacePointArray(vertexbuf);
		//		auto carray = m_obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type i = 0; i < varray.size(); i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				// IUV_̂ŁA̖ʂTJn
				searchUVisland_recursive(obj, fit.index(), foundUV, v2uvmap, selectedface);
				break;
			}
		}
	}
	return TRUE;
}

// O[vUV_
void UVToolsX::searchUVisland_recursive(cmq::CMQObject &obj, cmq::findex_t findex0,
	std::unordered_set<cmq::findex_t> &foundUV,
	const std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> &v2uvmap,
	std::unordered_set<cmq::findex_t> &selectedface)
{
	std::vector<int> vertexbuf(5);
	std::vector<cmq::CMQCoordinate> coordbuf(5);
	// T_ۑpXg
	std::list<cmq::findex_t> toexplore;
	toexplore.push_back(findex0);
	while(!toexplore.empty()) {
		cmq::findex_t findex = toexplore.front();
		toexplore.pop_front();
		foundUV.insert(findex);
		auto varray = obj.getFacePointArray(findex, vertexbuf);
		auto carray = obj.getFaceCoordinateArray(findex, coordbuf);
		// wʂ\eUV_ɂď
		for(std::vector<cmq::vindex_t>::size_type i = 0; i < varray.size(); i++) {
			// TʃZbgɑ}
			cmq::CMQCoordinate coord = carray[i];
			// _UV_
			auto findit = v2uvmap.equal_range(varray[i]);
			for(auto it = findit.first; it != findit.second; ++it) {
				auto flindex2 = it->second;
				std::pair<cmq::findex_t, int> fl2pair = cmq::flId2FaceLinePair(flindex2);
				// ʂTς݂ȂXLbv
				if(selectedface.find(fl2pair.first) != selectedface.end()) {
					continue;
				}
				auto carray2 = obj.getFaceCoordinateArray(fl2pair.first);
				cmq::CMQCoordinate coord2 = carray2[fl2pair.second];
				if(cmq::CMQCoordinate::uvcomp(coord, coord2)) {
					// UV_ŴŁAאUVʂł
					// ʂTςmapɑ}
					selectedface.insert(fl2pair.first);
					// TpXgɒǉ
					toexplore.push_back(fl2pair.first);
				}
			}
		}
	}
}

// UVOBJI
BOOL UVToolsX::selUV2OBJvert(MQDocument doc_)
{
	CursorKeeper cur;
	BOOL select = FALSE;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	std::vector<cmq::vindex_t> varray;
	varray.reserve(5);
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		varray.clear();
		int vcount = fit->getFacePointCount();
		for(int i = 0; i < vcount; i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				if(varray.empty()) {
					varray = obj.getFacePointArray(fit.index(), varray);
				}
				obj.addSelectVertex(varray[i]);
				select = TRUE;
			}
		}
	}
	return select;
}

BOOL UVToolsX::selUV2OBJline(MQDocument doc_)
{
	CursorKeeper cur;
	BOOL select = FALSE;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		int vcount = fit->getFacePointCount();
		for(int i = 0; i < vcount; i++) {
			if(doc.isSelectUVVertex(obj, fit.index(), i) && doc.isSelectUVVertex(obj, fit.index(), (i + 1) % vcount)) {
				doc.addSelectLine(obj, fit.index(), i);
				select = TRUE;
			}
		}
	}
	return select;
}

BOOL UVToolsX::selUV2OBJface(MQDocument doc_)
{
	CursorKeeper cur;
	BOOL select = FALSE;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		bool faceselected = true;
		int vcount = fit->getFacePointCount();
		// łIUV_炻̖ʂ͑IȂ
		for(int i = 0; i < vcount; i++) {
			if(!doc.isSelectUVVertex(obj, fit.index(), i)) {
				faceselected = false;
				break;
			}
		}
		if(faceselected) {
			fit->addSelect();
			select = TRUE;
		}
	}
	return select;
}

// OBJUVI
BOOL UVToolsX::selOBJUV2vert(MQDocument doc_)
{
	CursorKeeper cur;
	BOOL select = FALSE;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	std::vector<cmq::vindex_t> varray(5);
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		fit->getFacePointArray(varray);
		for(int i = 0; i < varray.size(); i++) {
			if(doc.isSelectVertex(obj, varray[i])) {
				obj.addSelectUVVertex(fit.index(), i);
				select = TRUE;
			}
		}
	}
	return select;
}

BOOL UVToolsX::selOBJUV2line(MQDocument doc_)
{
	CursorKeeper cur;
	BOOL select = FALSE;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		int vcount = fit->getFacePointCount();
		for(int i = 0; i < vcount; i++) {
			if(doc.isSelectLine(obj, fit.index(), i)) {
				obj.addSelectUVVertex(fit.index(), i);
				obj.addSelectUVVertex(fit.index(), (i + 1) % vcount);
				select = TRUE;
			}
		}
	}
	return select;
}

BOOL UVToolsX::selOBJUV2face(MQDocument doc_)
{
	CursorKeeper cur;
	BOOL select = FALSE;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		if(doc.isSelectFace(obj, fit.index())) {
			int vcount = fit->getFacePointCount();
			for(int i = 0; i < vcount; i++) {
				obj.addSelectUVVertex(fit.index(), i);
			}
			select = TRUE;
		}
	}
	return select;
}

// UV_͈̔(u,v̍őlAŏl)擾
void UVToolsX::getUVrange(cmq::CMQObject &obj, float &umin, float &umax, float &vmin, float &vmax)
{
	CursorKeeper cur;
	umin = vmin = 1.0f;
	umax = vmax = 0.0f;
	bool selected = false;
	std::vector<cmq::CMQCoordinate> coordbuf(5);
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto carray = obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type cindex = 0; cindex < carray.size(); cindex++) {
			if(obj.isSelectUVVertex(fit.index(), cindex)) {
				umin = min(umin, carray[cindex].u());
				umax = max(umax, carray[cindex].u());
				vmin = min(vmin, carray[cindex].v());
				vmax = max(vmax, carray[cindex].v());
				selected = true;
			}
		}
	}
	if(!selected) {
		umin = vmin = 0.0f;
		umax = vmax = 1.0f;
	}
}

// I𒸓_֐ڒ
BOOL UVToolsX::grabUVvertex(MQDocument doc_, bool faceselectedmode)
{
	return sewUV(doc_, faceselectedmode, true);
}

// 񒸓_LʂUV_Xg擾
void UVToolsX::getFaceFrom2Vert(cmq::CMQObject &obj, std::vector<std::pair<cmq::flindex_t, cmq::flindex_t>> &uvlist, cmq::vindex_t vindex1, cmq::vindex_t vindex2)
{
	// v1,v2̊eאږʃXg̐ϏW߂
	auto farray1 = obj.getVertexRelatedFaces(vindex1);
	std::sort(farray1.begin(), farray1.end());
	auto farray2 = obj.getVertexRelatedFaces(vindex2);
	std::sort(farray2.begin(), farray2.end());
	std::vector<cmq::findex_t> intersection;
	std::set_intersection(farray1.begin(), farray1.end(), farray2.begin(), farray2.end(), std::back_inserter(intersection));

	// eʂ̊YUV_I
	for(auto findex : intersection) {
		auto varray = obj.getFacePointArray(findex);
		// ʂ\UV_łIĂ̂grabΏۊOȂ̂ŏO
		bool selected = false;
		for(std::vector<cmq::vindex_t>::size_type j = 0; j < varray.size(); j++) {
			if(obj.isSelectUVVertex(findex, j)) {
				selected = true;
				break;
			}
		}
		if(selected) {
			continue;
		}
		for(std::vector<cmq::vindex_t>::size_type j = 0; j < varray.size(); j++) {
			cmq::vindex_t v1 = varray[j];
			cmq::vindex_t v2 = varray[(j + 1) % varray.size()];
			auto fl1 = cmq::makeFaceLineId(findex, j);
			auto fl2 = cmq::makeFaceLineId(findex, (j + 1) % varray.size());
			// [v1,v2]ƕ[vindex,vindex2]ӂł
			if(vindex1 == v2 && vindex2 == v1) {
				uvlist.push_back(std::make_pair(fl2, fl1));
				break;
			}
			else if(vindex1 == v1 && vindex2 == v2) {
				uvlist.push_back(std::make_pair(fl1, fl2));
				break;
			}
		}
	}
}

// Iӂ֖ʁAQڒ
BOOL UVToolsX::grabUVface(MQDocument doc_, bool islandmode, bool regular, bool notmove, bool faceselectedmode)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	std::vector<int> vertexbuf(5);
	std::vector<int> vertexbuf2(5);
	std::vector<cmq::CMQCoordinate> coordbuf(5);
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto varray = fit->getFacePointArray(vertexbuf);
		for(std::vector<cmq::vindex_t>::size_type lindex = 0; lindex < varray.size(); lindex++) {
			int lindex1 = (lindex + 1) % varray.size();
			if(obj.isSelectUVVertex(fit.index(), lindex) &&
				obj.isSelectUVVertex(fit.index(), lindex1)) {
				// IĂӂ\UV_ڒ̕ӂƂ
				auto carrayp = obj.getFaceCoordinateArray(fit.index(), coordbuf);
				// ړUVWl
				int lindex2 = (lindex + varray.size() - 1) % varray.size();
				cmq::CMQCoordinate &p0 = carrayp[lindex];
				cmq::CMQCoordinate &p1 = carrayp[lindex1];
				cmq::CMQCoordinate &p2 = carrayp[lindex2];
				MQCoordinate pvec = p1.getCoordinate() - p0.getCoordinate(); // ӃxNg
				MQCoordinate pface = p2.getCoordinate() - p0.getCoordinate(); // ʕsxNg
				float pnorm = pvec.u * pface.v - pvec.v * pface.u >= 0.0 ? 1.0f : -1.0f; // Wn̖@xNg
				if(pvec.u == 0 && pvec.v == 0) {
					// IUV_W0̕
					// ڒ̃xNgvZłȂ̂ŏȂ
					continue;
				}

				std::vector<std::pair<cmq::flindex_t, cmq::flindex_t>> uvlist;
				if(notmove) {
					// regularizê݂lindex1,2uvlistɓAړ̊_p0ɂ
					pvec = -pvec;
					p0 = p1;
					auto fl1 = cmq::makeFaceLineId(fit.index(), lindex);
					auto fl2 = cmq::makeFaceLineId(fit.index(), lindex1);
					uvlist.push_back(std::make_pair(fl2, fl1));
				}
				else {
					// grabgetFaceFrom2Vertuvlistɓ
					// ړΏۖʂiI2_܂ޖʁj
					cmq::vindex_t vp0 = varray[lindex];
					cmq::vindex_t vp1 = varray[lindex1];
					getFaceFrom2Vert(obj, uvlist, vp0, vp1);
				}
				for(auto line : uvlist) {
					auto fpair1 = cmq::flId2FaceLinePair(line.first);
					auto fpair2 = cmq::flId2FaceLinePair(line.second);
					cmq::findex_t findexq = fpair1.first; // ړ
					// ړΏۖʂ̃Xgɓo^
					std::unordered_set<cmq::findex_t> facelist;
					facelist.insert(findexq);
					if(islandmode) {
						// islandmodȅꍇAO[vʂǉ
						// _UV}bv
						std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> v2uvmap(makev2uvmap(obj));
						// ςݖ
						std::unordered_set<cmq::findex_t> selectedface;
						selectedface.reserve(obj.getFaceCount());
						// TJnʂ邽߂ɑIUV_
						selectedface.insert(findexq);
						searchUVisland_recursive(obj, findexq, facelist, v2uvmap, selectedface);
					}

					// Wϊs擾
					MQMatrix r;
					float rate = 1.0f;
					if(!regular) {
						// KȂ UVʏł̕ϊ
						MQCoordinate q0;
						r = getUVR(obj, r, q0, rate, findexq, fpair1.second, fpair2.second, pvec);
						// ړ
						for(cmq::findex_t findex : facelist) {
							if(faceselectedmode && !obj.isSelectFace(findex)) {
								continue;
							}
							movefaceuv(obj, findex, r, q0, p0, rate);
						}
					}
					else {
						// K 3Dԃ|SUVʂ֕ϊ
						MQPoint q0;
						for(cmq::findex_t findex : facelist) {
							if(faceselectedmode && !obj.isSelectFace(findex)) {
								continue;
							}
							auto targetarray = obj.getFacePointArray(findex);
							if(targetarray.size() > 3) {
								// lp`ȏ|S
								polygonmove(obj, findex, fpair1.second, fpair2.second, p0, pvec, pnorm);
							}
							else {
								// Op`|S
								r = get3DR(obj, r, q0, rate, findexq, fpair1.second, fpair2.second, pvec, pnorm);
								for(cmq::findex_t findex : facelist) {
									moveface3Duv(obj, findex, r, q0, p0, rate);
								}
							}
						}
					}
				}
			}
		}
	}
	return TRUE;
}

// TriangulateꂽOp|SXgtarrayאڃc[tree
void UVToolsX::TriangleTree::maketree(std::vector<cmq::TrianglePoints> &tarray, std::vector<TriangleTree> &tree)
{
	// c[
	for(int i = 0; i < tree.size(); i++) {
		tree[i].m_index = i; // indexݒ
	}
	for(int i = 0; i < tarray.size(); i++) {
		auto &tpoints = tarray[i];
		TriangleTree &node0 = tree[i];
		for(int k = 0; k < 3; k++) {
			node0.m_idx = tpoints;
//			node0.m_v[k] = tpoints[k]; // _CfbNXZbg
		}
		for(int j = 0; j < 3; j++) {
			if(node0.get(j)) {
				continue;
			}
			int idx1 = tpoints.idx[j];
			int idx2 = tpoints.idx[(j + 1) % 3];
			// Op|Siidx1,idx2ŗאڂOp|S
			auto pair = findneighbor(tarray, i, idx1, idx2);
			if(pair.first < 0) {
				// AEgC
				continue;
			}
			TriangleTree &node1 = tree[pair.first];
			node1.set(pair.second, &node0);
			node0.set(j, &node1);
		}
	}
}

// tarrayiԖڂ̖ʈȊOŒ_idx1,idx2܂ޖʂ
std::pair<int, int> UVToolsX::TriangleTree::findneighbor(std::vector<cmq::TrianglePoints> &tarray, int i, int idx1, int idx2)
{
	for(int k = 0; k < tarray.size(); k++) {
		if(k == i) {
			continue;
		}
		auto &tpoints = tarray[k];
		for(int l = 0; l < 3; l++) {
			if(tpoints.idx[l] == idx2 && tpoints.idx[(l + 1) % 3] == idx1) {
				std::pair<int, int> pair = std::make_pair(k, l);
				return pair;
			}
		}
	}
	return std::make_pair<int, int>(-1, -1);
}

// 3DԂ̎lp`ȏ̖findexlindex1,lindex2ŁAUVʂ̍Wp0̃xNgpeɐڂ悤ړ
void UVToolsX::polygonmove(cmq::CMQObject &obj, cmq::findex_t findex, int lindex1, int lindex2, cmq::CMQCoordinate &p0, MQCoordinate &pvec, float pnorm)
{
	// ڒWnuxNg
	float pveclen = GetSize(pvec);
	MQCoordinate pe = pvec / pveclen;

	// lp`ȏTriangulate
	auto varray = obj.getFacePointArray(findex);
	auto tarray = obj.triangulate(findex);
	// Opʂ̗אڃc[
	std::vector<TriangleTree> tree(tarray.size());
	TriangleTree::maketree(tarray, tree);

	cmq::CMQCoordinate coordinit(-1, MQCoordinate(-FLT_MAX, -FLT_MAX));
	std::vector<cmq::CMQCoordinate> newpoints(varray.size(), coordinit); // Wϊ̒_Wi[p

	// tree̒lindex1, lindex2܂ގOp|S
	auto pair = TriangleTree::findneighbor(tarray, -1, lindex1, lindex2);
	if(!(pair.first != -1 && pair.second != -1)) {
		// ʂŗאڂĂ
		return;
	}
	TriangleTree &t = tree[pair.first];
	int i = pair.second;
	func(obj, &t, varray, i, p0, pe, pveclen, pnorm, newpoints);

	// |Sړ
	obj.setFaceCoordinateArray(findex, newpoints);
}

// 3DԂ̃|SUVʂ֓]ʂϊs
MQMatrix makeR(MQMatrix &r, const MQPoint &qx, const MQPoint &qy, const MQPoint &qz, float pnorm, float rate, const MQCoordinate &pe)
{
	// qx,qy,qzxNg̃[JW[hWɕϊ
	MQMatrix r1;
	for(int i = 0; i < 16; i++) { r1.t[i] = 0.0f; }
	r1.SetColumn3(0, qx);
	r1.SetColumn3(1, -qy * pnorm);
	r1.SetColumn3(2, qz);
	r1._44 = 1.0f;
	// ratẽXP[
	MQMatrix rs;
	for(int i = 0; i < 16; i++) { rs.t[i] = 0.0f; }
	rs._11 = rs._22 = rs._33 = rate;
	rs._44 = 1.0f;
	// [hWpexNgUV[JWɕϊ
	MQMatrix r2;
	for(int i = 0; i < 16; i++) { r2.t[i] = 0.0f; }
	r2._11 = pe.u;
	r2._12 = pe.v;
	r2._21 = -pe.v;
	r2._22 = pe.u;
	r2._33 = pnorm;
	r2._44 = 1.0f;
	r = r1 * rs *r2;
	return r;
}

// Op|Sc[ԂUVʂɓWJċA
void UVToolsX::func(cmq::CMQObject &obj, TriangleTree *t, std::vector<cmq::vindex_t> &varray,
	int i, cmq::CMQCoordinate &p0, MQCoordinate &pe, float pveclen, float pnorm, std::vector<cmq::CMQCoordinate> &newpoints)
{
	int idx1 = t->m_idx.idx[i];
	int idx2 = t->m_idx.idx[(i + 1) % 3];
	int idx3 = t->m_idx.idx[(i + 2) % 3];
	assert(idx1 != -1);
	assert(idx2 != -1);
	assert(idx3 != -1);

	MQPoint q0 = obj.getVertex(varray[idx2]);
	MQPoint qvec = obj.getVertex(varray[idx1]) - q0;
	float qveclen = GetSize(qvec);
	MQPoint qe = qvec / qveclen; // ړWnXPʃxNg
	qe.normalize();
	float rate = pveclen / qveclen; // ڒ̈ړɑ΂g嗦
	// nꂽӂXƂĖʕYƖ@Z߂
	MQPoint qy = obj.getVertex(varray[idx3]) - q0;
	MQPoint qz = cmq::CMQPoint::outercross(qe, qy);
	qz.normalize();
	qy = cmq::CMQPoint::outercross(qz, qe);
	qy.normalize();

	// ϊs
	MQMatrix r;
	makeR(r, qe, qy, qz, pnorm, rate, pe);

	// ϊsgĎOp|S̊eWAUVWƂnewpointsɊi[
	for(int idx : t->m_idx.idx) {
		if(newpoints[idx].u() >= -100.0 && newpoints[idx].v() >= -100.0) {
			// UVvZς݂̔
			continue;
		}
		// ʂ̊eUV_̈ʒuϊ
		MQPoint p = obj.getVertex(varray[idx]);
		p -= q0;
		p = p * r;
		MQCoordinate coord(p.x + p0.u(), p.y + p0.v());
		newpoints[idx].getCoordinate() = coord;
	}
	t->m_done = true; // ς݃tO
	// אڎOp|SɂčċA
	for(int i = 0; i < 3; i++) {
		if(t->m_p[i] && !t->m_p[i]->m_done) {
			// p0, pe, pveclenO̎Op|S̍WŐݒ肵
			int idx1 = t->m_idx.idx[i];
			int idx2 = t->m_idx.idx[(i + 1) % 3];
			cmq::CMQCoordinate p0 = newpoints[idx1];
			cmq::CMQCoordinate p1 = newpoints[idx2];
			MQCoordinate pvec = p1.getCoordinate() - p0.getCoordinate();
			float pveclen = GetSize(pvec);
			MQCoordinate pe = pvec / pveclen; // ڒWnuxNg
			int thisindex = -1;
			for(int j = 0; j < 3; j++) {
				if(t->m_p[i]->m_p[j] == t) {
					thisindex = j;
					break;
				}
			}
			assert(thisindex != -1);
			func(obj, t->m_p[i], varray, thisindex, p0, pe, pveclen, pnorm, newpoints);
		}
	}
}

// UV|Sz
// findexʒuq0ʒup0ֈړArateŊgkāArŉ]
void UVToolsX::movefaceuv(cmq::CMQObject &obj, cmq::findex_t findex, MQMatrix &r, const MQCoordinate &q0, const cmq::CMQCoordinate &p0, float rate)
{
	auto carray = obj.getFaceCoordinateArray(findex);
	for(std::vector<cmq::CMQCoordinate>::size_type vindex = 0; vindex < carray.size(); vindex++) {
		// ʂ̊eUV_̈ʒuϊ
		cmq::CMQCoordinate &coord = carray[vindex];
		float u = coord.u() - q0.u;
		float v = coord.v() - q0.v;
		float u2 = (u * r._11 + v * r._21) * rate;
		float v2 = (u * r._12 + v * r._22) * rate;
		coord.u() = u2 + p0.u();
		coord.v() = v2 + p0.v();
	}
	// WϊʂZbg
	obj.setFaceCoordinateArray(findex, carray);
}

// ẑ߂̉]sr߂
MQMatrix UVToolsX::getUVR(cmq::CMQObject &obj, MQMatrix &r, MQCoordinate &q0, float &rate, cmq::findex_t findexq, int lindex1, int lindex2, const MQCoordinate &pvec)
{
	// ڒWnuxNg
	float pveclen = GetSize(pvec);
	MQCoordinate pe = pvec / pveclen;

	auto carrayq = obj.getFaceCoordinateArray(findexq);
	q0 = carrayq[lindex1].getCoordinate();
	cmq::CMQCoordinate q1 = carrayq[lindex2];
	MQCoordinate qvec = q1.getCoordinate() - q0; // ӃxNg
	float qveclen = GetSize(qvec);
	qveclen = qveclen == 0 ? 1.0 : qveclen;
	MQCoordinate qe = qvec / qveclen; // ړWnuxNg
	rate = pveclen / qveclen; // ڒ̈ړɑ΂g嗦
	// ϊs
	r._11 = pe.u * qe.u + pe.v * qe.v;
	r._12 = -pe.u * qe.v + pe.v * qe.u;
	r._21 = -pe.v * qe.u + pe.u * qe.v;
	r._22 = pe.v * qe.v + pe.u * qe.u;
	r._13 = r._14 = r._23 = r._24 = r._31 = r._32 = r._33 = r._34 = r._41 = r._42 = r._43 = r._44 = 0.0f;
	return r;
}

// Kċz
void UVToolsX::moveface3Duv(cmq::CMQObject &obj, cmq::findex_t findex, MQMatrix &r, const MQPoint &q0, const cmq::CMQCoordinate &p0, float rate)
{
	auto varray = obj.getFacePointArray(findex);
	std::vector<cmq::CMQCoordinate> carray(varray.size());
	for(std::vector<cmq::vindex_t>::size_type i = 0; i < varray.size(); i++) {
		// ʂ̊eUV_̈ʒuϊ
		MQPoint p = obj.getVertex(varray[i]);
		p -= q0;
		p = p * r;
		cmq::CMQCoordinate coord;
		coord.u() = p.x + p0.u();
		coord.v() = p.y + p0.v();
		carray[i] = coord;
	}
	// WϊʂZbg
	obj.setFaceCoordinateArray(findex, carray);
}

// Kċẑ߂̉]sr߂iOp|ŜݑΉŁj
// 3DԂ̃|SUVʂֈړ
MQMatrix UVToolsX::get3DR(cmq::CMQObject &obj, MQMatrix &r, MQPoint &q0, float &rate, cmq::findex_t findexq, int lindex1, int lindex2, const MQCoordinate &pvec, float pnorm)
{
	// ڒWnuxNg
	float pveclen = GetSize(pvec);
	MQCoordinate pe = pvec / pveclen;

	auto varray = obj.getFacePointArray(findexq);
	q0 = obj.getVertex(varray[lindex1]);
	MQPoint q1 = obj.getVertex(varray[lindex2]);
	MQPoint qvec = q1 - q0; // ӃxNg
	float qveclen = GetSize(qvec);
	qveclen = qveclen == 0 ? 1.0 : qveclen;
	MQPoint qe = qvec / qveclen; // ړWnXPʃxNg
	rate = pveclen / qveclen; // ڒ̈ړɑ΂g嗦
	MQPoint qz = obj.getNormalVector(findexq, lindex1);
	qz.normalize();
	MQPoint qy = cmq::CMQPoint::outercross(qz, qe);
	// ϊs
	makeR(r, qe, qy, qz, pnorm, rate, pe);
	return r;
}

BOOL UVToolsX::grab_regular(MQDocument doc_, bool faceselectedmode)
{
	return grabUVface(doc_, false, true, false, faceselectedmode);
}

// ]
BOOL UVToolsX::invertUVsel(MQDocument doc_, invert_direction_t direction)
{
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	// IUV_̍W͈
	float umin = 1.0f, umax = 0.0f, vmin = 1.0f, vmax = 0.0f;
	getUVrange(obj, umin, umax, vmin, vmax);
	return invertUV(doc_, umin, umax, vmin, vmax, direction);
}

// ]
BOOL UVToolsX::invertUV(MQDocument doc_, float umin, float umax, float vmin, float vmax, invert_direction_t direction)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();

	// _
	float umid = (umin + umax) / 2.0f;
	float vmid = (vmin + vmax) / 2.0f;

	std::vector<cmq::CMQCoordinate> coordbuf(5);
	// IUV_̍WύX
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto carray = obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type cindex = 0; cindex < carray.size(); cindex++) {
			if(obj.isSelectUVVertex(fit.index(), cindex)) {
				cmq::CMQCoordinate &coord = carray[cindex];
				switch(direction) {
				case INVERT_HORIZONTAL: coord.u() = umid * 2 - coord.u(); break;
				case INVERT_VERTICAL: coord.v() = vmid * 2 - coord.v(); break;
				}
			}
		}
		obj.setFaceCoordinateArray(fit.index(), carray);
	}
	return TRUE;
}

// qL
BOOL UVToolsX::turnUVsel(MQDocument doc_, turn_direction_t direction)
{
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	// IUV_̍W͈
	float umin = 1.0f, umax = 0.0f, vmin = 1.0f, vmax = 0.0f;
	getUVrange(obj, umin, umax, vmin, vmax);
	return turnUV(doc_, umin, umax, vmin, vmax, direction);
}

// qL
BOOL UVToolsX::turnUV(MQDocument doc_, float umin, float umax, float vmin, float vmax, turn_direction_t direction)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	std::vector<cmq::CMQCoordinate> coordbuf(5);
	// IUV_̍WύX
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto carray = obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type i = 0; i < carray.size(); i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				cmq::CMQCoordinate &coord = carray[i];
				switch(direction) {
				case TURN_UP: coord.v() = vmin * 2.0f - coord.v(); break;
				case TURN_DOWN: coord.v() = vmax * 2.0f - coord.v(); break;
				case TURN_LEFT: coord.u() = umin * 2.0f - coord.u(); break;
				case TURN_RIGHT: coord.u() = umax * 2.0f - coord.u(); break;
				}
			}
		}
		obj.setFaceCoordinateArray(fit.index(), carray);
	}
	return TRUE;
}

// ]
BOOL UVToolsX::rotate(MQDocument doc_, const MQCoordinate &pivot, float angle)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();

	float cosvalue = cos(angle * PI / 180.0f);
	float sinvalue = sin(angle * PI / 180.0f);

	std::vector<cmq::CMQCoordinate> coordbuf(5);
	// IUV_̍WύX
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto carray = obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type i = 0; i < carray.size(); i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				cmq::CMQCoordinate &coord = carray[i];
				MQCoordinate c = coord.getCoordinate() - pivot;
				float u = c.u * cosvalue - c.v * sinvalue;
				float v = c.u * sinvalue + c.v * cosvalue;
				c.u = u;
				c.v = v;
				coord.set(c + pivot);
			}
		}
		obj.setFaceCoordinateArray(fit.index(), carray);
	}
	return TRUE;
}

// ړ
BOOL UVToolsX::move(MQDocument doc_, const MQCoordinate &delta)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	std::vector<cmq::CMQCoordinate> coordbuf(5);
	// IUV_̍WύX
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto carray = obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type i = 0; i < carray.size(); i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				cmq::CMQCoordinate &coord = carray[i];
				MQCoordinate c = coord.getCoordinate();
				coord.set(c + delta);
			}
		}
		obj.setFaceCoordinateArray(fit.index(), carray);
	}
	return TRUE;
}

// gk
BOOL UVToolsX::scale(MQDocument doc_, const MQCoordinate &pivot, float rate)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	std::vector<cmq::CMQCoordinate> coordbuf(5);
	// IUV_̍WύX
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto carray = obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type i = 0; i < carray.size(); i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				cmq::CMQCoordinate &coord = carray[i];
				MQCoordinate c = coord.getCoordinate() - pivot;
				c = c * rate;
				coord.set(c + pivot);
			}
		}
		obj.setFaceCoordinateArray(fit.index(), carray);
	}
	return TRUE;
}

// DAڒ
BOOL UVToolsX::sewUV(MQDocument doc_, bool faceselectedmode, bool grabmode)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	// _UV}bv
	std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> v2uvmap(makev2uvmap(obj));
	// ςݒ_
	std::unordered_set<cmq::vindex_t> processedvertex;
	processedvertex.reserve(obj.getVertexCount());
	std::vector<cmq::CMQCoordinate> coordbuf(5);
	std::vector<cmq::CMQCoordinate> coordbuf2(5);
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto varray = fit->getFacePointArray();
		auto carray = obj.getFaceCoordinateArray(fit.index(), coordbuf);
		for(std::vector<cmq::CMQCoordinate>::size_type i = 0; i < varray.size(); i++) {
			if(obj.isSelectUVVertex(fit.index(), i)) {
				// IUV_̒_
				cmq::vindex_t vindex = varray[i];
				if(processedvertex.find(vindex) != processedvertex.end()) {
					// ς
					continue;
				}
				auto findit = v2uvmap.equal_range(vindex);
				cmq::CMQCoordinate coordsum;
				int count = 0;
				bool skip = false;
				for(auto it = findit.first; it != findit.second; ++it) {
					auto flpair = cmq::flId2FaceLinePair(it->second);
//					if(faceselectedmode && !obj.isSelectFace(flpair.first)) {
//						// IʂȂ̂őΏۊO
//						continue;
//					}
					auto carray1 = obj.getFaceCoordinateArray(flpair.first, coordbuf2);
					if(obj.isSelectUVVertex(it->second)) {
						auto flpair = cmq::lineId2VertexPair(it->second);
						auto coord = carray1[flpair.second];
						if(grabmode && !cmq::CMQCoordinate::uvcomp(carray[i], coord)) {
							// _̑IUViUVʒüقȂUV_IĂj
							skip = true;
							break;
						}
						coordsum.getCoordinate() += coord.getCoordinate();
						count++;
					}
				}
				if(count > 1) {
					if(grabmode && skip) {
						// ڒ[h͑IUV_̒_͑ΏۊO
						continue;
					}
					// _̑IUV_ȏ゠ꍇ͕ψʒuɂ
					coordsum.getCoordinate() /= float(count);
				}
				for(auto it = findit.first; it != findit.second; ++it) {
					auto flpair = cmq::flId2FaceLinePair(it->second);
					if(faceselectedmode && !obj.isSelectFace(flpair.first)) {
						// IʂȂ̂őΏۊO
						continue;
					}
					if(!grabmode && !obj.isSelectUVVertex(it->second)) {
						// D[h͑IUV_̂ݖD
						continue;
					}
					auto flindex = it->second;
					auto flpair2 = cmq::flId2FaceLinePair(flindex);
					// ʂɊ܂܂eUV̍WύX
					obj.setFaceCoordinate(flpair2.first, flpair2.second, coordsum);
				}
				processedvertex.insert(vindex);
			}
		}
	}
	return TRUE;
}

BOOL UVToolsX::regularize(MQDocument doc_, bool faceselectedmode)
{
	return grabUVface(doc_, false, true, true, faceselectedmode);
}

BOOL UVToolsX::gather(MQDocument doc_, bool widemode, float delta)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	// _UV}bv
	std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> v2uvmap(makev2uvmap(obj));
	float delta2 = (float)pow(delta, 2.0);

	cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(&obj);
	for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
		auto findit = v2uvmap.equal_range(vit.index());
		std::vector<MQCoordinate> coordarry;
		for(auto it = findit.first; it != findit.second; ++it) {
			auto flpair = cmq::flId2FaceLinePair(it->second);
			if(!obj.isSelectUVVertex(flpair.first, flpair.second)) {
				continue;
			}
			auto carray = obj.getFaceCoordinateArray(flpair.first);
			auto &coord = carray[flpair.second];
			for(int i = 0; i < coordarry.size(); i++) {
				auto c2 = coordarry[i];
				auto diff = pow(coord.u() - c2.u, 2.0) + pow(coord.v() - c2.v, 2.0);
				if(diff > 0.0 && diff < delta2) {
					coord.set(c2);
					obj.setFaceCoordinateArray(flpair.first, carray);
					break;
				}
			}
			coordarry.push_back(coord.getCoordinate());
		}
	}
	return TRUE;
}

std::unordered_map<int, int> selectAsymmetricVertex(cmq::CMQDocument &doc, cmq::CMQObject &obj)
{
	float delta = 0.001f;
	int vcount = obj.getVertexCount();
	bool is_selected = false; // NɑI𒸓_邩
	for(int i = 0; i < vcount; i++) {
		if(obj.getVertexRefCount(i) == 0) {
			continue;
		}
		is_selected |= doc.isSelectVertex(obj, i);
	}

	std::unordered_map<int, int> symmetric; // Ώ̒_yAi[
//	std::unordered_set<int> asymmetric; // Ώ̒_݂Ȃ̂i[
	for(int i = 0; i < vcount; i++) {
		if(obj.getVertexRefCount(i) == 0) {
			continue;
		}
		if(is_selected && !doc.isSelectVertex(obj, i)) {
			continue;
		}
		if(symmetric.find(i) != symmetric.end()) {
			// Ɍς
			continue;
		}
		MQPoint pi = obj.getVertex(i);
		if(abs(pi.x) < delta) {
			// ΏۋEʏ
			symmetric.insert(std::make_pair(i, i));
			doc.deleteSelectVertex(obj, i);
			continue;
		}
		pi.x = -pi.x;
		bool is_symmetric = false;
		for(int j = 0; j < vcount; j++) {
			if(obj.getVertexRefCount(j) == 0) {
				continue;
			}
			if(i == j) {
				continue;
			}
			MQPoint pj = obj.getVertex(j);
			if((pi - pj).norm() < delta * delta) {
				// Ώے_
				is_symmetric = true;
				symmetric.insert(std::make_pair(j, i));
				symmetric.insert(std::make_pair(i, j));
				doc.deleteSelectVertex(obj, i);
				doc.deleteSelectVertex(obj, j);
				break;
			}
		}
	}
	return symmetric;
}

// vertexes̊e_̑Ώ̒_
std::vector<int> getSymmetricVertexesofFace(const std::vector<int> &vertexes, const std::unordered_map<int, int> &vsymmetric)
{
	std::vector<int> symvertexes; // vertex1̊e_̑Ώۈʒu̒_
	symvertexes.reserve(vertexes.size());
	for(int v : vertexes) {
		auto it = vsymmetric.find(v);
		if(it != vsymmetric.end()) {
			symvertexes.push_back(it->second);
		}
		else {
			break;
		}
	}
	return symvertexes;
}

// _Qvertexesō\ʂ
int findFace(cmq::CMQObject &obj, const std::vector<int> &vertexes)
{
	std::set<int> target; // e_ɋʂ̖ʂ邽߂̐ϏWi[
	for(int v2 : vertexes) {
		std::vector<int> faces = obj.getVertexRelatedFaces(v2);
		std::set<int> s; // v2܂ޖʂsetBtargetƂ̐ϏW߂邽
		for(int findex : faces) {
			s.insert(findex);
		}
		if(target.empty()) {
			target = s;
		}
		else {
			std::set<int> tmp;
			set_intersection(target.begin(), target.end(), s.begin(), s.end(), inserter(tmp, tmp.begin()));
			if(tmp.empty()) {
				// ʂĂȂ
				target.clear();
				break;
			}
			target = tmp;
		}
	}
	if(target.empty()) {
		// Ώے_͈ʂʂĂȂ
		return -1;
	}
	return *target.begin();
}


// _tɓ
bool isReverseVector(const std::vector<int> &vector1, const std::vector<int> &vector2)
{
	if(vector1.front() == vector2.back()) {
		return equal(vector2.begin(), vector2.end(), vector1.rbegin());
	}
	else {
		auto it = find(vector1.begin(), vector1.end(), vector2.back());
		if(it == vector1.end()) {
			// _ĂȂ
			return false;
		}
		std::vector<int> vector3(vector1.size());
		rotate_copy(vector1.begin(), it, vector1.end(), vector3.begin());
		return equal(vector2.begin(), vector2.end(), vector3.rbegin());
	}
}

BOOL UVToolsX::symmetrize(MQDocument doc_, bool widemode)
{
	CursorKeeper cur;
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj = doc.getCurrentObject();
	std::unordered_map<int, int> vsymmetric = selectAsymmetricVertex(doc, obj);
	std::unordered_map<int, int> fsymmetric; // Ώ̖ʂyAi[

	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		if(!obj.isSelectFace(fit.index())) {
			continue;
		}
		if(!obj.getFaceVisible(fit.index())) {
			// \
			continue;
		}
		if(fsymmetric.find(fit.index()) != fsymmetric.end()) {
			// Ɍς
			continue;
		}
		//bool is_symmetric = false;
		// findex̑Ώۖʂ = findexɊ܂܂e_̑Ώ̒_ō\ʂ
		auto vertexes = obj.getFacePointArray(fit.index());
		auto symvertexes = getSymmetricVertexesofFace(vertexes, vsymmetric);
		if(vertexes.size() != symvertexes.size()) {
			// ȂΏے_
			continue;
		}
		// _Qvertexesō\ʂ
		int fsymindex = findFace(obj, symvertexes);
		if(fsymindex == -1) {
			continue;
		}
		if(fsymindex == fit.index()) {
			// g񂾑Ώ̂Ȗʂłꍇ
			continue;
		}
		// Ώ̖ʂUVlRs[
		cmq::vindex_t vindex0 = vertexes[0];
		auto vfindit = vsymmetric.find(vindex0);
		if(vfindit == vsymmetric.end()) {
			// something wrong
			continue;
		}
		cmq::vindex_t vindex0sym = vfindit->second;
		int indexdif = -1;
		auto varraysym = obj.getFacePointArray(fsymindex);
		auto carraysym = obj.getFaceCoordinateArray(fsymindex);
		for(int i = 0; i < carraysym.size(); i++) {
			if(varraysym[i] == vindex0sym) {
				indexdif = i;
				break;
			}
		}
		assert(indexdif != -1);
		std::vector<cmq::CMQCoordinate> carray(carraysym.size());
		for(int i = 0; i < carraysym.size(); i++) {
			cmq::CMQCoordinate csym = carraysym[(carraysym.size() - i + indexdif) % carraysym.size()];
			if(widemode) {
				// wide mode
				csym.u() =1.0f - csym.u();
			}
			carray[i] = csym;
		}
		obj.setFaceCoordinateArray(fit.index(), carray);
	}
	return TRUE;
}

BOOL UVToolsX::setMem(MQDocument doc_, int index)
{
	CursorKeeper cur;
	if(index < 0 || index >= STORESIZE) {
		return FALSE;
	}
	m_store[index].clear();
	cmq::CMQDocument doc(doc_);
	int ocount = doc.getObjectCount();
	bool selected = false;
	for(int oindex = 0; oindex < ocount; oindex++) {
		cmq::CMQObject obj = doc.getObject(oindex);
		UINT uid = obj.getUniqueID();
		if(!obj.isLive()) {
			continue;
		}
		if(obj.getLocking() || !obj.getVisible()) {
			continue;
		}
		std::unordered_set<cmq::flindex_t> selected_set;
		selected_set.reserve(obj.getFaceCount());
		cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
		for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
			int vcount = fit->getFacePointCount();
			for(int i = 0; i < vcount; i++) {
				if(obj.isSelectUVVertex(fit.index(), i)) {
					UINT uid = obj.getFaceUniqueID(fit.index());
					cmq::flindex_t flindex = cmq::makeFaceLineId(uid, i);
					selected_set.insert(flindex);
					selected = true;
				}
			}
		}
		if(!selected_set.empty()) {
			m_store[index].insert(std::make_pair(uid, selected_set));
		}
	}
	return selected ? TRUE : FALSE;
}

BOOL UVToolsX::restoreMem(MQDocument doc_, int index)
{
	CursorKeeper cur;
	if(index < 0 || index >= STORESIZE) {
		return FALSE;
	}
	if(m_store[index].empty()) {
		return FALSE;
	}
	cmq::CMQDocument doc(doc_);
	int ocount = doc.getObjectCount();
	bool selected = false;
	for(auto setpair : m_store[index]) {
		UINT uid = setpair.first;
		cmq::CMQObject obj = doc.getObjectFromUniqueID(uid);
		if(!obj.isLive()) {
			continue;
		}
		auto selected_set = setpair.second;
		for(cmq::flindex_t flindex : selected_set) {
			auto fpair = cmq::flId2FaceLinePair(flindex);
			UINT uid = fpair.first;
			cmq::findex_t findex = obj.getFaceIndexFromUniqueID(uid);
			obj.addSelectUVVertex(cmq::makeFaceLineId(findex, fpair.second));
			selected = true;
		}
	}
	return selected ? TRUE : FALSE;
}

bool UVToolsX::locksetselected(MQDocument doc_, int index, bool isfix, bool isappend)
{
	if(!m_bActivate) {
		return false;
	}
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj(doc.getCurrentObject());
	UINT ouid = obj.getUniqueID();
	auto &map = m_lock[index];
	if(isfix) {
		if(!isappend) {
			map.clear();
		}
		// I𒸓_W
		auto findit = map.find(ouid);
		if(findit == map.end()) {
			std::unordered_map<cmq::flindex_t, MQCoordinate> newlist;
			auto it = map.insert(std::make_pair(ouid, newlist));
			findit = it.first;
		}
		auto &list = findit->second;
		cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
		for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
			auto carray = obj.getFaceCoordinateArray(fit.index());
			for(int i = 0; i < carray.size(); i++) {
				if(obj.isSelectUVVertex(fit.index(), i)) {
					UINT fuid = obj.getFaceUniqueID(fit.index());
					cmq::flindex_t flindex = cmq::makeFaceLineId(fuid, i);
					MQCoordinate coord = carray[i].getCoordinate();
					list.insert(std::make_pair(flindex, coord));
				}
			}
		}
		if(list.empty()) {
			map.erase(findit);
		}
	}
	m_isfix[index] = !map.empty();
	return m_isfix[index];
}

void UVToolsX::lockclear(MQDocument doc, int index)
{
	m_lock[index].clear();
	m_isfix[index] = false;
}

bool UVToolsX::lock(MQDocument doc_, int index, bool isfix)
{
	isfix = isfix && !m_lock[index].empty();
	return m_isfix[index] = isfix;
}

void UVToolsX::OnEndDocument(MQDocument doc)
{
	for(int i = 0; i < STORESIZE; i++) {
		m_lock[i].clear();
		m_isfix[i] = false;
	}
	m_Window->clearbutton();
}

void UVToolsX::OnDraw(MQDocument doc_, MQScene scene, int width, int height)
{
	if(!m_bActivate) {
		return;
	}
	cmq::CMQDocument doc(doc_);
	cmq::CMQObject obj(doc.getCurrentObject());
	for(int i = 0; i < STORESIZE; i++) {
		if(m_isfix[i]) {
			UINT ouid = obj.getUniqueID();
			auto findit = m_lock[i].find(ouid);
			if(findit == m_lock[i].end()) {
				continue;
			}
			auto &list = findit->second;
			for(auto it = list.begin(); it != list.end(); ++it) {
				auto fpair = cmq::flId2FaceLinePair(it->first);
				UINT fuid = fpair.first;
				cmq::findex_t findex = obj.getFaceIndexFromUniqueID(fuid);
				if(findex != -1) {
					auto carray = obj.getFaceCoordinateArray(findex);
					int lindex = fpair.second;
					if(lindex >= carray.size() || lindex < 0) {
						continue;
					}
					carray[lindex].set(it->second);
					obj.setFaceCoordinateArray(findex, carray);
				}
			}
		}
	}
}


//----------------------------------------------------------------------------------------------

// wfindex̖ʂ̑SUV_ɋLĂ邩
bool UVToolsX::memorycheck(cmq::CMQObject &obj, DWORD memory, cmq::findex_t findex)
{
	UINT fuid = obj.getFaceUniqueID(findex);
	for(int i = 0; i < STORESIZE; i++) {
		if(!((1 << i) & memory)) {
			continue;
		}
		auto findit = m_store[i].find(obj.getUniqueID());
		if(findit == m_store[i].end()) {
			continue;
		}
		std::unordered_set<cmq::flindex_t> &memory = findit->second;
		int vcount = obj.getFacePointCount(findex);
		if(vcount < 2) {
			continue;
		}
		bool selected = true;
		for(int j = 0; j < vcount; j++) {
			cmq::flindex_t flindex = cmq::makeFaceLineId(fuid, j);
			if(memory.find(flindex) == memory.end()) {
				selected = false;
				break;
			}
		}
		if(selected) {
			return true;
		}
	}
	return false;
}


// |S̕ӂ`
BOOL UVToolsX::writewire(cmq::CMQDocument &doc, Gdiplus::Bitmap *bitmap, long width, long height, bool bcurobjflg, bool bcurmatflg, DWORD mode, DWORD memory)
{
	Gdiplus::Pen pen(Gdiplus::Color(255, 255, 255, 255), 1.0);
//	std::unique_ptr<Gdiplus::Graphics> graphis(Gdiplus::Graphics::FromImage(bitmap));
	Gdiplus::Graphics* graphics = Gdiplus::Graphics::FromImage(bitmap);

	cmq::CMQObject curobj(doc.getCurrentObject());
	cmq::mindex_t curmindex = doc.getCurrentMaterialIndex();
	MQMaterial curmat = doc.getMaterial(curmindex);
	std::string curtexture = curmat ? doc.getTextureName(curmat) : "";

	for(cmq::oindex_t oindex = 0; oindex < doc.getObjectCount(); oindex++) {
		cmq::CMQObject obj(doc.getObject(oindex));
		if(!obj.isLive() || !obj.getVisible()) {
			continue;
		}
		if(bcurobjflg && obj != curobj) {
			continue;
		}
		cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
		cmq::mindex_t curmindex = doc.getCurrentMaterialIndex();
		std::vector<Gdiplus::Point> points;
		points.reserve(5);
		for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
			auto mindex = obj.getFaceMaterial(fit.index());
			if(bcurmatflg && mindex != curmindex) {
				continue;
			}
			MQMaterial mat = doc.getMaterial(mindex);
			if(mat && doc.getTextureName(mat) != curtexture) {
				continue;
			}
			if(memory) {
				if(!memorycheck(obj, memory, fit.index())) {
					continue;
				}
			}
			auto carray = obj.getFaceCoordinateArray(fit.index());
			points.resize(carray.size());
			for(int i = 0; i < carray.size(); i++) {
				points[i].X = (INT)(width * carray[i].u());
				points[i].Y = (INT)(height * carray[i].v());
			}
			//setPoint(points[carray.size()], width * carray[0].u(), height * carray[0].v());
			Gdiplus::GraphicsPath path;
			path.AddPolygon(&points.front(), points.size());
			graphics->DrawPath(&pen, &path);
		}
	}
	delete graphics;
	return TRUE;
}

// 1UVʂO[[VF[fBO`
void GouraudShading(Gdiplus::Bitmap *pbitmap, const std::vector<cmq::CMQCoordinate> &carray, const std::vector<MQColor> &colors, long width, long height)
{
	// UV|Svł݂āAul̍ŏl/őlƁAcolorvectorɊi[
	int index0 = -1;
	int index1 = -1;
	double vmin = FLT_MAX;
	double vmax = -FLT_MAX;
	for(int i = 0; i < carray.size(); i++) {
		if(carray[i].v() < vmin) { index0 = i;  vmin = carray[i].v(); }
		if(carray[i].v() > vmax) { index1 = i;  vmax = carray[i].v(); }
	}
	double fheight = height - 0.1;
	double fwidth = width - 0.1;
	long y0 = (long)round(vmin * fheight);
	long y1 = (long)round(vmax * fheight);
	std::vector<long> x0(y1 - y0 + 1);
	std::vector<long> x1(y1 - y0 + 1);
	std::vector<MQColor> col0(y1 - y0 + 1);
	std::vector<MQColor> col1(y1 - y0 + 1);
	for(int i = index0; i != index1; i = NEXTIDX(i, carray.size())) {
		// vŏ㒸_E
		int nextidx = NEXTIDX(i, carray.size());
		long yfrom = (long)round(carray[i].v() * fheight);
		long yto = (long)round(carray[nextidx].v() * fheight);
		long xfrom = (long)round(carray[i].u() * fwidth);
		long xto = (long)round(carray[nextidx].u() * fwidth);
		MQColor colfrom = colors[i];
		MQColor colto = colors[nextidx];
		long ylen = yto - yfrom + 1;
		for(int j = 0; j < ylen; j++) {
			col0[yfrom - y0 + j] = colfrom * ((float)(ylen - j) / ylen) + colto * ((float)j / ylen);
			x0[yfrom - y0 + j] = (long)round(xfrom * ((float)(ylen - j) / ylen) + xto * ((float)j / ylen));
		}
	}
	for(int i = index0; i != index1; i = PREVIDX(i, carray.size())) {
		// vŏ㒸_獶
		int previdx = PREVIDX(i, carray.size());
		long yfrom = (long)round(carray[i].v() * fheight);
		long yto = (long)round(carray[previdx].v() * fheight);
		long xfrom = (long)round(carray[i].u() * fwidth);
		long xto = (long)round(carray[previdx].u() * fwidth);
		MQColor colfrom = colors[i];
		MQColor colto = colors[previdx];
		long ylen = yto - yfrom + 1;
		for(int j = 0; j <= yto - yfrom; j++) {
			col1[yfrom - y0 + j] = colfrom * ((float)(ylen - j) / ylen) + colto * ((float)j / ylen);
			x1[yfrom - y0 + j] = (long)round(xfrom * ((float)(ylen - j) / ylen) + xto * ((float)j / ylen));
		}
	}
	// ey(v)WɁAx(u)Wŏƍő̊Ԃ̃sNZɂāAF`ԂĖ߂
	for(long y = 0; y <= y1 - y0; y++) {
		long xfrom, xto;	
		MQColor coll, colh;
		if(x0[y] < x1[y]) {
			xfrom = x0[y];
			xto = x1[y];
			coll = col0[y];
			colh = col1[y];
		}
		else {
			xfrom = x1[y];
			xto = x0[y];
			coll = col1[y];
			colh = col0[y];
		}
		long xlen = xto - xfrom + 1;
		for(int x = 0; x < xlen; x++) {
			MQColor col = coll * ((float)(xlen - x) / xlen) + colh * ((float)x / xlen);
			Gdiplus::Color color(255, (BYTE)(col.r * 255), (BYTE)(col.g * 255), (BYTE)(col.b * 255));
			pbitmap->SetPixel(xfrom + x, y0 + y, color);
		}
	}
}

// eʂɂāAאږʂƂ̃X[WOꍇ́AX[WOpxȉłΖʖ@̕ςZbg
// łȂΖʖ@̂̂Zbg
bool UVToolsX::calcshade(cmq::CMQDocument &doc, cmq::CMQObject &obj, std::unordered_map<cmq::flindex_t, MQColor> &shademap)
{
	// 
	MQScene scene = doc.getScene(0);
	int lightnum = scene->GetGlobalDirectionalLightNum();
	std::vector<std::pair<MQPoint, MQColor>> veclight(lightnum);
	for(int i = 0; i < lightnum; i++) {
		MQPoint point = scene->GetGlobalDirectionalLightDirection(i);
		MQColor color = scene->GetGlobalDirectionalLightColor(i);
		veclight[i] = std::make_pair(point, color);
	}
	// ʖ@
	std::unordered_map<cmq::findex_t, MQPoint> normalmap;
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		MQPoint normal = obj.getNormalVector(fit.index());
		normal.normalize();
		normalmap.insert(std::make_pair(fit.index(), normal));
	}

	float smooth = obj.getSmoothAngle();
	float cossmooth = cos(2.0f * PI * smooth / 360.0f);
	// AevZ
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		// ʂ̊e_ɂ@߂
		auto findit = normalmap.find(fit.index());
		if(findit == normalmap.end()) {
			// something wrong
			continue;
		}
		MQPoint normal0 = findit->second;
		cmq::mindex_t mindex = obj.getFaceMaterial(fit.index());
		MQMaterial mat = nullptr;
		if(mindex >= 0) {
			mat = doc.getMaterial(mindex);
		}
		auto varray = obj.getFacePointArray(fit.index());
		for(int i = 0; i < varray.size(); i++) {
			MQPoint normal = normal0;
			if(obj.getShading()) {
				// X[WO
				int count = 1;
				auto faces = obj.getVertexRelatedFaces(varray[i]);
				for(auto findex : faces) {
					if(findex == fit.index()) {
						continue;
					}
					auto findit2 = normalmap.find(findex);
					if(findit2 == normalmap.end()) {
						// shomthing wrong
						continue;
					}
					MQPoint n = findit2->second;
					if(cmq::CMQPoint::innercross(normal0, n) >= cossmooth) {
						// X[WOpxȏȂ瑫킹
						normal += n;
						count++;
					}
				}
				normal /= (float)count;
			}

			// @ƌ炱̒_ɂAe߂
			MQColor col(0, 0, 0);
			for(auto light : veclight) {
				auto vec = light.first;
				auto lightcol = light.second;
				MQColor ambient = mat ? mat->GetAmbient() : 0.5f;
				float cross = -cmq::CMQPoint::innercross(normal, vec);
				MQColor diffuse = (cross < 0.0f) ? 0.0f : (cross * (mat ? mat->GetDiffuse() : 0.5f));
				col = col + lightcol * (0.3f * ambient + 0.7f * diffuse); // ambient:diffse3:7(ĂƂ)ō
				if(col.r > 1.0) { col.r = 1.0; }
				if(col.g > 1.0) { col.g = 1.0; }
				if(col.b > 1.0) { col.b = 1.0; }
			}
			cmq::flindex_t flindex = cmq::makeFaceLineId(fit.index(), i);
			shademap.insert(std::make_pair(flindex, col));
		}
	}
	return true;
}

// UVʂ`
BOOL UVToolsX::writevertex(cmq::CMQDocument &doc, Gdiplus::Bitmap *pbitmap, long width, long height, bool bcurobjflg, bool bcurmatflg, DWORD mode, DWORD memory)
{
	cmq::CMQObject curobj(doc.getCurrentObject());
	cmq::mindex_t curmindex = doc.getCurrentMaterialIndex();
	MQMaterial curmat = doc.getMaterial(curmindex);
	std::string curtexture = curmat ? doc.getTextureName(curmat) : "";

	for(cmq::oindex_t oindex = 0; oindex < doc.getObjectCount(); oindex++) {
		cmq::CMQObject obj(doc.getObject(oindex));
		if(!obj.isLive() || !obj.getVisible()) {
			continue;
		}
		if(bcurobjflg && obj != curobj) {
			continue;
		}
		bool shadeflag = false;
		cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
		for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
			auto mindex = obj.getFaceMaterial(fit.index());
			if(bcurmatflg && mindex != curmindex) {
				continue;
			}
			MQMaterial mat = doc.getMaterial(mindex);
			if(mat && doc.getTextureName(mat) != curtexture) {
				continue;
			}
			shadeflag = true;
		}
		if(!shadeflag) {
			continue;
		}
		std::unordered_map<cmq::flindex_t, MQColor> shademap;
		if(mode & TEX_SHADER) {
			if(!calcshade(doc, obj, shademap)) {
				return FALSE;
			}
		}
		for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
			auto mindex = obj.getFaceMaterial(fit.index());
			if(bcurmatflg && mindex != curmindex) {
				continue;
			}
			MQMaterial mat = doc.getMaterial(mindex);
			if(mat && doc.getTextureName(mat) != curtexture) {
				continue;
			}
			if(memory) {
				if(!memorycheck(obj, memory, fit.index())) {
					continue;
				}
			}
			auto carray = obj.getFaceCoordinateArray(fit.index());
			std::vector<MQColor> colors(carray.size());
			if(mode & CLIPALPHA) {
				for(int i = 0; i < carray.size(); i++) {
					colors[i] = MQColor(1.0f, 1.0f, 1.0f);
				}
			}
			else if(mode & TEX_SHADER) {
				for(int i = 0; i < carray.size(); i++) {
					cmq::flindex_t flindex = cmq::makeFaceLineId(fit.index(), i);
					auto findit = shademap.find(flindex);
					if(findit == shademap.end()) {
						continue;
					}
					MQColor c = findit->second;
					colors[i] = MQColor(c.r, c.g, c.b);
				}
			}
			else if(mode & TEX_VERT) {
				for(int i = 0; i < carray.size(); i++) {
					DWORD c = obj.m_obj->GetFaceVertexColor(fit.index(), i);
					colors[i] = MQColor((float)GetRValue(c) / 255.0f, (float)GetGValue(c) / 255.0f, (float)GetBValue(c) / 255.0f);
				}
			}
			GouraudShading(pbitmap, carray, colors, width, height);
		}
	}
	return FALSE;
}

// UVʂPF`
BOOL UVToolsX::writemat(cmq::CMQDocument &doc, Gdiplus::Bitmap *bitmap, long width, long height, bool bcurobjflg, bool bcurmatflg, DWORD mode, DWORD memory, bool ismat)
{
	std::unique_ptr<Gdiplus::Graphics> graphics(Gdiplus::Graphics::FromImage(bitmap));

	cmq::CMQObject curobj(doc.getCurrentObject());
	cmq::mindex_t curmindex = doc.getCurrentMaterialIndex();
	MQMaterial curmat = doc.getMaterial(curmindex);
	std::string curtexture = curmat ? doc.getTextureName(curmat) : "";

	for(cmq::oindex_t oindex = 0; oindex < doc.getObjectCount(); oindex++) {
		cmq::CMQObject obj(doc.getObject(oindex));
		if(!obj.isLive() || !obj.getVisible()) {
			continue;
		}
		if(bcurobjflg && obj != curobj) {
			continue;
		}
		MQColor objc = obj.m_obj->GetColor();
		auto objcolor = Gdiplus::Color::MakeARGB(255, (BYTE)(objc.r * 255), (BYTE)(objc.g * 255), (BYTE)(objc.b * 255));
		float fheight = (float)height - 0.1f;
		float fwidth = (float)width - 0.1f;
		cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
		for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
			auto mindex = obj.getFaceMaterial(fit.index());
			if(bcurmatflg && mindex != curmindex) {
				continue;
			}
			MQMaterial mat = doc.getMaterial(mindex);
			if(mat && doc.getTextureName(mat) != curtexture) {
				continue;
			}
			if(memory) {
				if(!memorycheck(obj, memory, fit.index())) {
					continue;
				}
			}
			auto carray = obj.getFaceCoordinateArray(fit.index());
			std::vector<Gdiplus::Point> points;
			for(int i = 0; i < carray.size(); i++) {
				Gdiplus::Point p;
				p.X = (INT)round(carray[i].u() * fwidth);
				p.Y = (INT)round(carray[i].v() * fheight);
				points.push_back(p);
			}
			Gdiplus::Color color;
			if(mode & CLIPALPHA) {
				color = Gdiplus::Color::MakeARGB(255, 255, 255, 255);
			}
			else if(ismat) {
				MQColor c = mat->GetColor();
				color = Gdiplus::Color::MakeARGB(255, (BYTE)(c.r * 255), (BYTE)(c.g * 255), (BYTE)(c.b * 255));
			}
			else {
				color = objcolor;
			}
			Gdiplus::GraphicsPath path;
			path.AddPolygon(&points.front(), points.size());
			Gdiplus::SolidBrush brush1(color);
			graphics->FillPath(&brush1, &path);
		}
	}
	return TRUE;
}

bool GetEncoderClsid(const wchar_t *format, CLSID* pClsid)
{
	UINT numEncoders = 0;
	UINT size = 0;
	Gdiplus::GetImageEncodersSize(&numEncoders, &size);
	if(size == 0) {
		return false;
	}
	std::vector<Gdiplus::ImageCodecInfo> vecImageCodecInfo(size);
	GetImageEncoders(numEncoders, size, &vecImageCodecInfo.front());
	vecImageCodecInfo.resize(numEncoders);
	for(auto info : vecImageCodecInfo) {
		if(wcscmp(format, info.MimeType) == 0) {
			*pClsid = info.Clsid;
			return true;
		}
	}
	return false;
}

void UVToolsX::writedata(MQDocument doc_, long width, long height, const std::wstring &filename, DWORD mode, DWORD memory, filetype_t filetype)
{
	std::unique_ptr<Gdiplus::Bitmap> bitmap(new Gdiplus::Bitmap(width, height));

	// rbg}bv`
	cmq::CMQDocument doc(doc_);
	bool bcurobjflg = !!(mode & TEX_CUROBJ);
	bool bcurmatflg = !!(mode & TEX_CURMAT);
	if(mode & TEX_OBJ) {
		writemat(doc, bitmap.get(), width, height, bcurobjflg, bcurmatflg, mode, memory, false);
	}
	if(mode & TEX_MAT) {
		writemat(doc, bitmap.get(), width, height, bcurobjflg, bcurmatflg, mode, memory, true);
	}
	if(mode & TEX_VERT) {
		writevertex(doc, bitmap.get(), width, height, bcurobjflg, bcurmatflg, mode, memory);
	}
	if(mode & TEX_SHADER) {
		writevertex(doc, bitmap.get(), width, height, bcurobjflg, bcurmatflg, mode, memory);
	}
	if(mode & TEX_WIRE) {
		writewire(doc, bitmap.get(), width, height, bcurobjflg, bcurmatflg, mode, memory);
	}

	switch(filetype) {
	case FILE_BMP:
	{
		CLSID encoderParam = GUID_NULL;
		if(GetEncoderClsid(L"image/bmp", &encoderParam)) {
			const wchar_t *wfilename = filename.c_str();
			auto status = bitmap->Save(wfilename, &encoderParam);
		}
	}
		break;
	case FILE_JPG:
	{
		CLSID encoderParam = GUID_NULL;
		if(GetEncoderClsid(L"image/jpeg", &encoderParam)) {
			const wchar_t *wfilename = filename.c_str();
			auto status = bitmap->Save(wfilename, &encoderParam);
		}
	}
		break;
	case FILE_PNG:
	{
		CLSID encoderParam = GUID_NULL;
		if(GetEncoderClsid(L"image/png", &encoderParam)) {
			const wchar_t *wfilename = filename.c_str();
			auto status = bitmap->Save(wfilename, &encoderParam);
		}
	}
		break;
	case FILE_TGA:
	{
		// TGAwb_
		TGAFILEHEADER tgafh;
		ZeroMemory(&tgafh, sizeof(tgafh));
		tgafh.imagetype = 2; // Uncompressed, True-color Image
		tgafh.imagemap.imageoriginX = 0;
		tgafh.imagemap.imageoriginY = 0;
		tgafh.imagemap.imagewidth = (WORD)width;
		tgafh.imagemap.imageheight = (WORD)height;
		tgafh.imagemap.bitperpixel = 32;
		tgafh.imagemap.descripter = 8;

		TGAFILEFOOTER tgaff;
		ZeroMemory(&tgaff, sizeof(tgaff));
		strcpy_s(tgaff.signature, "TRUEVISION-XFILE.");

		int imagesize = width * height * sizeof(DWORD);

		bitmap->RotateFlip(Gdiplus::RotateNoneFlipY);
		Gdiplus::BitmapData bitmapdata;
		Gdiplus::Rect rect(0, 0, width, height);
		Gdiplus::Status status = bitmap->LockBits(&rect, Gdiplus::ImageLockModeRead, PixelFormat32bppARGB, &bitmapdata);
		if(status == Gdiplus::Ok) {
			// TGAo
			DWORD len = 0;
			bitmapdata.Stride = -1;
			HANDLE hfile = CreateFile(filename.c_str(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
			if(hfile != INVALID_HANDLE_VALUE) {
				WriteFile(hfile, &tgafh, sizeof(tgafh), &len, NULL);
				WriteFile(hfile, bitmapdata.Scan0, imagesize, &len, NULL);
				WriteFile(hfile, &tgaff, sizeof(tgaff), &len, NULL);
				CloseHandle(hfile);
			}
			bitmap->UnlockBits(&bitmapdata);
		}
	}
		break;
	case FILE_CLIPBOARD:
	{
		int imagesize = width * height * sizeof(DWORD);

		HBITMAP hBitmap = nullptr;
		EmptyClipboard();
		if(bitmap->GetHBITMAP(Gdiplus::Color(255, 0, 0, 0), &hBitmap) == Gdiplus::Ok) {
			DIBSECTION dibs;
			GetObject(hBitmap, sizeof(dibs), &dibs);
			dibs.dsBmih.biCompression = BI_RGB;
			HDC hDC = GetDC(NULL);
			HBITMAP hDDB = CreateDIBitmap(hDC, &dibs.dsBmih, CBM_INIT, dibs.dsBm.bmBits, (BITMAPINFO*)&dibs.dsBmih, DIB_RGB_COLORS);
			ReleaseDC(NULL, hDC);
			if(OpenClipboard(nullptr)) {
				EmptyClipboard();
				HANDLE hHandle = ::SetClipboardData(CF_BITMAP, hDDB);
				if(hHandle) {
					OutputDebugString(L"ok");
				}
			}
			CloseClipboard();
		}
		DeleteObject(hBitmap);
	}
		break;
	default:
		// do nothing
		;
	}
}

// eNX`t@Co
BOOL UVToolsX::writefile(MQDocument doc_, int width, int height, const std::wstring &filename, DWORD mode, DWORD memory)
{
	CursorKeeper cur;
	filetype_t filetype = FILE_NONE;

	if(mode & CLIPCOLOR || mode & CLIPALPHA) {
		filetype = FILE_CLIPBOARD;
	}
	else {
		m_texturewidth = width;
		m_textureheight = height;
		m_texturemode = mode;
		m_filename = filename;
		auto pos = m_filename.rfind('.');
		if(pos != std::wstring::npos) {
			std::wstring ext = m_filename.substr(pos + 1);
			for(auto &c : ext) { c = tolower(c); }
			if(ext == L"bmp") {
				filetype = FILE_BMP;
			}
			else if(ext == L"jpg" || ext == L"jpeg") {
				filetype = FILE_JPG;
			}
			else if(ext == L"png") {
				filetype = FILE_PNG;
			}
			else if(ext == L"tga") {
				filetype = FILE_TGA;
			}
		}
	}

	Gdiplus::GdiplusStartupInput gdiSI;
	ULONG_PTR gdiToken;
	if(Gdiplus::GdiplusStartup(&gdiToken, &gdiSI, NULL) == Gdiplus::Ok) {
		writedata(doc_, width, height, filename, mode, memory, filetype);
		Gdiplus::GdiplusShutdown(gdiToken);
	}
	return TRUE;
}
