#include <windows.h>

#include "UVToolsXWindow.h"
#include "UVToolsX.h"

UVToolsXWindow::UVToolsXWindow(UVToolsX *plugin, MQWindowBase& parent)
: MQWindow(parent), m_pPlugin(plugin)
{
	SetTitle(L"UVToolsX");
	SetCloseButton(false);
	SetCanResize(false);
	SetInSpace(0.0);
	SetOutSpace(0.0);

	{
		MQGroupBox *group6 = CreateGroupBox(this, L"UV -> Object");
		group6->SetCanMinimize(true);
		group6->SetOutSpace(0.0);
		group6->SetInSpace(0.0);
		group6->AddResizeEvent(this, &UVToolsXWindow::resize);
		group6->SetAlignment(MQFrameBase::MQFrameAlignment::ALIGN_HORIZONTAL);
		group6->SetMatrixColumn(3);
		MQButton *m_button_uvobj_vert = CreateButton(group6, L"V");
		m_button_uvobj_vert->AddClickEvent(this, &UVToolsXWindow::selUV2OBJvert);
		MQButton *m_button_uvobj_line = CreateButton(group6, L"L");
		m_button_uvobj_line->AddClickEvent(this, &UVToolsXWindow::selUV2OBJline);
		MQButton *m_button_uvobj_face = CreateButton(group6, L"F");
		m_button_uvobj_face->AddClickEvent(this, &UVToolsXWindow::selUV2OBJface);
	}
	{
		MQGroupBox *group7 = CreateGroupBox(this, L"Object -> UV");
		group7->SetCanMinimize(true);
		group7->SetOutSpace(0.0);
		group7->SetInSpace(0.0);
		group7->AddResizeEvent(this, &UVToolsXWindow::resize);
		group7->SetAlignment(MQFrameBase::MQFrameAlignment::ALIGN_HORIZONTAL);
		group7->SetMatrixColumn(3);
		MQButton *m_button_objuv_vert = CreateButton(group7, L"V");
		m_button_objuv_vert->AddClickEvent(this, &UVToolsXWindow::selOBJ2UVvert);
		MQButton *m_button_objuv_line = CreateButton(group7, L"L");
		m_button_objuv_line->AddClickEvent(this, &UVToolsXWindow::selOBJ2UVline);
		MQButton *m_button_objuv_face = CreateButton(group7, L"F");
		m_button_objuv_face->AddClickEvent(this, &UVToolsXWindow::selOBJ2UVface);
	}
	{
		MQGroupBox *group1 = CreateGroupBox(this, L"Select");
		group1->SetCanMinimize(true);
		group1->SetOutSpace(0.0);
		group1->SetInSpace(0.0);
		group1->AddResizeEvent(this, &UVToolsXWindow::resize);
		MQButton *button1 = CreateButton(group1, L"floating");
		button1->AddClickEvent(this, &UVToolsXWindow::sel_floating);
		MQButton *button2 = CreateButton(group1, L"group");
		button2->AddClickEvent(this, &UVToolsXWindow::sel_group);
		MQButton *button3 = CreateButton(group1, L"same vertex");
		button3->AddClickEvent(this, &UVToolsXWindow::sel_samevert);
	}
	{
		MQGroupBox *group2 = CreateGroupBox(this, L"Grab");
		group2->SetCanMinimize(true);
		group2->SetOutSpace(0.0);
		group2->SetInSpace(0.0);
		group2->AddResizeEvent(this, &UVToolsXWindow::resize);
		{
			MQFrame *frame2_1 = CreateHorizontalFrame(group2);
			button_grab_vert = CreateButton(frame2_1, L"V");
			button_grab_vert->AddClickEvent(this, &UVToolsXWindow::grab_vert);
			button_grab_face = CreateButton(frame2_1, L"F");
			button_grab_face->AddClickEvent(this, &UVToolsXWindow::grab_face);
			button_grab_group = CreateButton(frame2_1, L"G");
			button_grab_group->AddClickEvent(this, &UVToolsXWindow::grab_face);
			button_grab_regular = CreateButton(frame2_1, L"R");
			button_grab_regular->AddClickEvent(this, &UVToolsXWindow::grab_regular);
			MQButton *button_sew = CreateButton(frame2_1, L"Sew");
			button_sew->AddClickEvent(this, &UVToolsXWindow::sew);
		}
		{
			MQFrame *frame2_2 = CreateHorizontalFrame(group2);
			MQButton *button_gather = CreateButton(frame2_2, L"Gather");
			button_gather->AddClickEvent(this, &UVToolsXWindow::gather);
			m_spin_symmetirize_delta = CreateDoubleSpinBox(frame2_2);
			m_spin_symmetirize_delta->SetVisibleColumn(6);
			m_spin_symmetirize_delta->SetDecimalDigit(4);
			m_spin_symmetirize_delta->SetPosition(0.001);
			m_spin_symmetirize_delta->SetMin(0.0);
			m_spin_symmetirize_delta->SetMax(1.0);
			m_spin_symmetirize_delta->SetIncrement(0.0002);
			m_check_selectedfacemode = CreateCheckBox(frame2_2, L"selected");
		}
	}
	{
		MQGroupBox *group3 = CreateGroupBox(this, L"Invert");
		group3->SetCanMinimize(true);
		group3->SetOutSpace(0.0);
		group3->SetInSpace(0.0);
		group3->AddResizeEvent(this, &UVToolsXWindow::resize);
		group3->SetAlignment(MQFrameBase::MQFrameAlignment::ALIGN_HORIZONTAL);
		group3->SetMatrixColumn(2);
		m_button_invert_h = CreateButton(group3, L"Horizontal");
		m_button_invert_h->AddClickEvent(this, &UVToolsXWindow::invert);
		m_button_invert_v = CreateButton(group3, L"Vertical");
		m_button_invert_v->AddClickEvent(this, &UVToolsXWindow::invert);
	}
	{
		MQGroupBox *group4 = CreateGroupBox(this, L"Trun");
		group4->SetCanMinimize(true);
		group4->SetOutSpace(0.0);
		group4->SetInSpace(0.0);
		group4->AddResizeEvent(this, &UVToolsXWindow::resize);
		group4->SetAlignment(MQFrameBase::MQFrameAlignment::ALIGN_HORIZONTAL);
		group4->SetMatrixColumn(4);
		m_button_turn_up = CreateButton(group4, L"U");
		m_button_turn_up->AddClickEvent(this, &UVToolsXWindow::turn);
		m_button_turn_down = CreateButton(group4, L"D");
		m_button_turn_down->AddClickEvent(this, &UVToolsXWindow::turn);
		m_button_turn_left = CreateButton(group4, L"L");
		m_button_turn_left->AddClickEvent(this, &UVToolsXWindow::turn);
		m_button_turn_right = CreateButton(group4, L"R");
		m_button_turn_right->AddClickEvent(this, &UVToolsXWindow::turn);
	}
	{
		MQGroupBox *group7 = CreateGroupBox(this, L"Manipulate");
		group7->SetCanMinimize(true);
		group7->SetOutSpace(0.0);
		group7->SetInSpace(0.0);
		MQButton *m_button_regularize = CreateButton(group7, L"Regularize");
		m_button_regularize->AddClickEvent(this, &UVToolsXWindow::regularize);
		MQFrame *frame7_1 = CreateHorizontalFrame(group7);
		MQButton *m_button_symmetrize = CreateButton(frame7_1, L"Symmetrize");
		m_button_symmetrize->AddClickEvent(this, &UVToolsXWindow::symmetrize);
		MQLabel *symmetrize_mode = CreateLabel(frame7_1, L"wide");
		m_check_symmetrize_mode = CreateCheckBox(frame7_1);
		MQButton *button_numerical = CreateButton(group7, L"Numerical");
		button_numerical->AddClickEvent(this, &UVToolsXWindow::numerical);
	}
	{
		MQGroupBox *group5 = CreateGroupBox(this, L"Memory");
		group5->SetCanMinimize(true);
		group5->SetOutSpace(0.0);
		group5->SetInSpace(0.0);
		group5->AddResizeEvent(this, &UVToolsXWindow::resize);
		group5->SetOutSpace(0.0);
		group5->SetInSpace(0.0);
		group5->SetAlignment(MQFrameBase::MQFrameAlignment::ALIGN_HORIZONTAL);
		group5->SetMatrixColumn(5);

		MQLabel*store_title = CreateLabel(group5, L"Store");
		for(int i = 0; i < UVToolsX::STORESIZE; i++) {
			wchar_t no[2];
			wsprintf(no, L"%d", i + 1);
			MQButton *p_button = CreateButton(group5, no);
			m_button_store[0].push_back(p_button);
			p_button->AddClickEvent(this, &UVToolsXWindow::set);
		}
		MQLabel*select = CreateLabel(group5, L"Restore");
		for(int i = 0; i < UVToolsX::STORESIZE; i++) {
			wchar_t no[2];
			wsprintf(no, L"%d", i + 1);
			MQButton *p_button = CreateButton(group5, no);
			m_button_store[1].push_back(p_button);
			p_button->AddClickEvent(this, &UVToolsXWindow::restore);
		}
		MQLabel*output = CreateLabel(group5, L"Lock");
		for(int i = 0; i < UVToolsX::STORESIZE; i++) {
			wchar_t no[2];
			wsprintf(no, L"%d", i + 1);
			MQButton *p_button = CreateButton(group5, no);
			m_button_lock.push_back(p_button);
			p_button->AddLeftDownEvent(this, &UVToolsXWindow::lock);
		}
	}
	{
		MQGroupBox *group6 = CreateGroupBox(this, L"File");
		group6->SetCanMinimize(true);
		group6->SetOutSpace(0.0);
		group6->SetInSpace(0.0);
		group6->AddResizeEvent(this, &UVToolsXWindow::resize);
		MQButton *button_file = CreateButton(group6, L"Output");
		button_file->AddClickEvent(this, &UVToolsXWindow::textureoutput);
	}

	AddHideEvent(this, &UVToolsXWindow::OnHide);
}

UVToolsXWindow::~UVToolsXWindow()
{
	obj_.deleteThis();
}

BOOL UVToolsXWindow::sel_floating(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selectFloatingUV(doc);
}

BOOL UVToolsXWindow::sel_group(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selectGroupUV(doc);
}

BOOL UVToolsXWindow::sel_samevert(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selectUVsame(doc);
}

BOOL UVToolsXWindow::selUV2OBJvert(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selUV2OBJvert(doc);
}

BOOL UVToolsXWindow::selUV2OBJline(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selUV2OBJline(doc);
}

BOOL UVToolsXWindow::selUV2OBJface(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selUV2OBJface(doc);
}

BOOL UVToolsXWindow::selOBJ2UVvert(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selOBJUV2vert(doc);
}

BOOL UVToolsXWindow::selOBJ2UVline(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selOBJUV2line(doc);
}

BOOL UVToolsXWindow::selOBJ2UVface(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->selOBJUV2face(doc);
}

BOOL UVToolsXWindow::grab_vert(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->grabUVvertex(doc, m_check_selectedfacemode->GetChecked());
}

BOOL UVToolsXWindow::grab_face(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->grabUVface(doc, sender == button_grab_group, false, false, m_check_selectedfacemode->GetChecked());
}

BOOL UVToolsXWindow::grab_regular(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->grab_regular(doc, m_check_selectedfacemode->GetChecked());
}

BOOL UVToolsXWindow::symmetrize(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->symmetrize(doc, m_check_symmetrize_mode->GetChecked());
}

BOOL UVToolsXWindow::set(MQWidgetBase *sender, MQDocument doc)
{
	auto findit = std::find(m_button_store[0].begin(), m_button_store[0].end(), sender);
	MQButton *button = dynamic_cast<MQButton*>(sender);
	_ASSERT(button);
	if(button->GetDown()) {
		MQDialog::DIALOG_RESULT result = MQDialog::MessageOkCancelBox(*this, L"Overwrite?", L"UVToolsX");
		if(result != MQDialog::DIALOG_OK) {
			return FALSE;
		}
	}
	if(findit != m_button_store[0].end()) {
		int index = (findit - m_button_store[0].begin());
		BOOL selected = m_pPlugin->setMem(doc, index);
		(*findit)->SetDown(!!selected);
	}
	return FALSE;
}

BOOL UVToolsXWindow::restore(MQWidgetBase *sender, MQDocument doc)
{
	auto findit = std::find(m_button_store[1].begin(), m_button_store[1].end(), sender);
	if(findit != m_button_store[1].end()) {
		int index = (findit - m_button_store[1].begin());
		return m_pPlugin->restoreMem(doc, index);
	}
	return FALSE;
}

BOOL UVToolsXWindow::lock(MQWidgetBase *sender, MQDocument doc, MQWidgetMouseParam& param)
{
	auto findit = std::find(m_button_lock.begin(), m_button_lock.end(), sender);
	if(findit != m_button_lock.end()) {
		int index = (findit - m_button_lock.begin());
		if(param.Ctrl) {
			bool isfix = m_pPlugin->locksetselected(doc, index, true, true);
			(*findit)->SetDown(isfix);
		}
		else if(param.Shift) {
			m_pPlugin->lockclear(doc, index);
			(*findit)->SetDown(false);
		}
		else {
			bool isfix = (*findit)->GetDown();
			isfix = m_pPlugin->lock(doc, index, !isfix);
			(*findit)->SetDown(isfix);
		}
	}
	return FALSE;
}

BOOL UVToolsXWindow::invert(MQWidgetBase *sender, MQDocument doc)
{
	UVToolsX::invert_direction_t direction = UVToolsX::INVERT_NO;
	if(sender == m_button_invert_h) {
		direction = UVToolsX::INVERT_HORIZONTAL;
	}
	else {
		direction = UVToolsX::INVERT_VERTICAL;
	}
	return m_pPlugin->invertUVsel(doc, direction);
}

BOOL UVToolsXWindow::turn(MQWidgetBase *sender, MQDocument doc)
{
	UVToolsX::turn_direction_t direction = UVToolsX::TURN_NO;
	if(sender == m_button_turn_up) { direction = UVToolsX::TURN_UP; }
	else if(sender == m_button_turn_down) { direction = UVToolsX::TURN_DOWN; }
	else if(sender == m_button_turn_left) { direction = UVToolsX::TURN_LEFT; }
	else if(sender == m_button_turn_right) { direction = UVToolsX::TURN_RIGHT; }
	return m_pPlugin->turnUVsel(doc, direction);
}

BOOL UVToolsXWindow::sew(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->sewUV(doc, false);
}

BOOL UVToolsXWindow::gather(MQWidgetBase *sender, MQDocument doc)
{
	float delta = (float)m_spin_symmetirize_delta->GetPosition();
	return m_pPlugin->gather(doc, false, delta);
}

BOOL UVToolsXWindow::regularize(MQWidgetBase *sender, MQDocument doc)
{
	return m_pPlugin->regularize(doc, m_check_selectedfacemode->GetChecked());
}

BOOL UVToolsXWindow::textureoutput(MQWidgetBase *sender, MQDocument doc)
{
	// file setting window
	TextureOutputDialog dlg = TextureOutputDialog(*this, m_pPlugin);
	dlg.m_spinwidth->SetPosition(m_pPlugin->m_texturewidth);
	dlg.m_spinheight->SetPosition(m_pPlugin->m_textureheight);
	dlg.m_editfilename->SetText(m_pPlugin->m_filename);
	if(dlg.Execute() != MQDialog::DIALOG_OK) {
		return FALSE;
	}

	// output
	std::wstring filename = dlg.m_editfilename->GetText();
	if(filename.empty()) {
		return FALSE;
	}
	DWORD mode = (dlg.m_checkwire->GetChecked() ? UVToolsX::TEX_WIRE : 0) |
		(dlg.m_checkvertexcol && dlg.m_checkvertexcol->GetChecked() ? UVToolsX::TEX_VERT : 0) |
		(dlg.m_checkmatcol->GetChecked() ? UVToolsX::TEX_MAT : 0) |
		(dlg.m_checkobjcol->GetChecked() ? UVToolsX::TEX_OBJ : 0) |
		(dlg.m_checkshader->GetChecked() ? UVToolsX::TEX_SHADER : 0) |
		(dlg.m_currentobj->GetChecked() ? UVToolsX::TEX_CUROBJ : 0) |
		(dlg.m_currentmat->GetChecked() ? UVToolsX::TEX_CURMAT : 0);
	int width = dlg.m_spinwidth->GetPosition();
	int height = dlg.m_spinheight->GetPosition();
	DWORD memory = 0;
	if(dlg.m_checkmemory->GetChecked()) {
		for(int i = 0; i < UVToolsX::STORESIZE; i++) {
			if(dlg.m_checkmemorystore[i]->GetChecked()) {
				memory |= (1 << i);
			}
		}
	}
	return m_pPlugin->writefile(doc, width, height, filename, mode, memory);
}

BOOL UVToolsXWindow::OnHide(MQWidgetBase *sender, MQDocument doc)
{
	if(m_pPlugin->m_bActivate) {
		m_pPlugin->m_bActivate = false;
		m_pPlugin->WindowClose();
	}
	return FALSE;
}

void UVToolsXWindow::resize()
{
	POINT size = GetJustSize();
	SetWidth(size.x);
	SetHeight(size.y);
}

void UVToolsXWindow::clearbutton()
{
	for(int i = 0; i < UVToolsX::STORESIZE; i++) {
		m_button_lock[i]->SetDown(false);
	}
}

//--------------------------------------------------

BOOL UVToolsXWindow::numerical(MQWidgetBase *sender, MQDocument doc)
{
	NumericalManipulateDialog dlg(*this, m_pPlugin, doc);
	if(dlg.Execute() != MQDialog::DIALOG_OK) {
		return FALSE;
	}
	//	return m_pPlugin->numerical(doc);
	return FALSE;
}

NumericalManipulateDialog::NumericalManipulateDialog(MQWindowBase& parent, UVToolsX *plugin, MQDocument doc) : MQDialog(parent), m_pPlugin(plugin), m_doc(doc)
{
	setlocale(LC_CTYPE, "");
	SetTitle(L"Numerical Manipulate");
	this->SetOutSpace(0.5);
	this->SetInSpace(0.5);

	{
		MQFrame *mainframe = CreateVerticalFrame(this);
		{
			MQFrame *frame1 = CreateHorizontalFrame(mainframe);
			MQFrame *frame1_0 = CreateVerticalFrame(frame1);
			frame1_0->SetFillBeforeRate(0.4);
			frame1_0->SetFillAfterRate(0.4);
			frame1_0->SetBackColor(0, 0, 0, 255);
			MQFrame *frame1_1 = CreateHorizontalFrame(frame1_0);
			MQLabel *label = CreateLabel(frame1_1, L"V min");
			MQDoubleSpinBox *spin = CreateDoubleSpinBox(frame1_1);
			m_spin[0] = spin;
			spin->SetMin(0.0);
			spin->SetMax(1.0);
			spin->SetIncrement(0.01);
			spin->SetVisibleColumn(6);
			spin->AddChangingEvent(this, &NumericalManipulateDialog::change_spin);
			MQFrame *frame1_2 = CreateHorizontalFrame(frame1_0);
			MQButton *button_init = CreateButton(frame1_2, L"Init");
			button_init->AddClickEvent(this, &NumericalManipulateDialog::click_init);
			MQButton *button_0 = CreateButton(frame1_2, L"0");
			button_0->AddClickEvent(this, &NumericalManipulateDialog::click_min);
			MQButton *button_mid = CreateButton(frame1_2, L"0.5");
			button_mid->AddClickEvent(this, &NumericalManipulateDialog::click_mid);
			MQButton *button_max = CreateButton(frame1_2, L"max");
			button_max->AddClickEvent(this, &NumericalManipulateDialog::click_max);
			m_button_init[0] = button_init;
			m_button_min[0] = button_0;
			m_button_mid[0] = button_mid;
			m_button_max[0] = button_max;
		}
		MQFrame *frame1B = CreateHorizontalFrame(mainframe);
		{
			MQFrame *frame2 = CreateVerticalFrame(frame1B);
			frame2->SetBackColor(0, 0, 0, 255);
			MQFrame *frame2_1 = CreateHorizontalFrame(frame2);
			MQLabel *label = CreateLabel(frame2_1, L"U min");
			MQDoubleSpinBox *spin = CreateDoubleSpinBox(frame2_1);
			m_spin[1] = spin;
			spin->SetMin(0.0);
			spin->SetMax(1.0);
			spin->SetIncrement(0.01);
			spin->SetVisibleColumn(6);
			spin->AddChangingEvent(this, &NumericalManipulateDialog::change_spin);
			MQFrame *frame2_2 = CreateHorizontalFrame(frame2);
			MQButton *button_init = CreateButton(frame2_2, L"Init");
			button_init->AddClickEvent(this, &NumericalManipulateDialog::click_init);
			MQButton *button_0 = CreateButton(frame2_2, L"0");
			button_0->AddClickEvent(this, &NumericalManipulateDialog::click_min);
			MQButton *button_mid = CreateButton(frame2_2, L"0.5");
			button_mid->AddClickEvent(this, &NumericalManipulateDialog::click_mid);
			MQButton *button_max = CreateButton(frame2_2, L"max");
			button_max->AddClickEvent(this, &NumericalManipulateDialog::click_max);
			m_button_init[1] = button_init;
			m_button_min[1] = button_0;
			m_button_mid[1] = button_mid;
			m_button_max[1] = button_max;
		}
		{
			MQFrame *frame5 = CreateHorizontalFrame(frame1B);
			frame5->SetBackColor(0, 0, 0, 255);
			{
				MQFrame *frame5_1 = CreateVerticalFrame(frame5);
				MQFrame *frame5_1_1 = CreateHorizontalFrame(frame5_1);
				MQLabel *labelu = CreateLabel(frame5_1_1, L"pivot U");
				MQDoubleSpinBox *spinu = CreateDoubleSpinBox(frame5_1_1);
				m_spin[4] = spinu;
				spinu->SetMin(0.0);
				spinu->SetMax(1.0);
				spinu->SetIncrement(0.01);
				spinu->SetVisibleColumn(6);
				spinu->AddChangingEvent(this, &NumericalManipulateDialog::change_pivot);
				MQFrame *frame5_1_2 = CreateHorizontalFrame(frame5_1);
				MQButton *button_init = CreateButton(frame5_1_2, L"Init");
				button_init->AddClickEvent(this, &NumericalManipulateDialog::click_init);
				MQButton *button_umin = CreateButton(frame5_1_2, L"min");
				button_umin->AddClickEvent(this, &NumericalManipulateDialog::click_min);
				MQButton *button_mid = CreateButton(frame5_1_2, L"0.5");
				button_mid->AddClickEvent(this, &NumericalManipulateDialog::click_mid);
				MQButton *button_umax = CreateButton(frame5_1_2, L"max");
				button_umax->AddClickEvent(this, &NumericalManipulateDialog::click_max);
				m_button_init[4] = button_init;
				m_button_min[4] = button_umin;
				m_button_mid[4] = button_mid;
				m_button_max[4] = button_umax;
			}
			{
				MQFrame *frame5_2 = CreateVerticalFrame(frame5);
				MQFrame *frame5_2_1 = CreateHorizontalFrame(frame5_2);
				MQLabel *labelv = CreateLabel(frame5_2_1, L"V");
				MQDoubleSpinBox *spinv = CreateDoubleSpinBox(frame5_2_1);
				m_spin[5] = spinv;
				spinv->SetMin(0.0);
				spinv->SetMax(1.0);
				spinv->SetIncrement(0.01);
				spinv->SetVisibleColumn(6);
				spinv->AddChangingEvent(this, &NumericalManipulateDialog::change_pivot);
				MQFrame *frame5_2_2 = CreateHorizontalFrame(frame5_2);
				MQButton *button_init = CreateButton(frame5_2_2, L"Init");
				button_init->AddClickEvent(this, &NumericalManipulateDialog::click_init);
				MQButton *button_vmin = CreateButton(frame5_2_2, L"min");
				button_vmin->AddClickEvent(this, &NumericalManipulateDialog::click_min);
				MQButton *button_mid = CreateButton(frame5_2_2, L"0.5");
				button_mid->AddClickEvent(this, &NumericalManipulateDialog::click_mid);
				MQButton *button_vmax = CreateButton(frame5_2_2, L"max");
				button_vmax->AddClickEvent(this, &NumericalManipulateDialog::click_max);
				m_button_init[5] = button_init;
				m_button_min[5] = button_vmin;
				m_button_mid[5] = button_mid;
				m_button_max[5] = button_vmax;
			}
		}
		{
			MQFrame *frame3 = CreateVerticalFrame(frame1B);
			frame3->SetBackColor(0, 0, 0, 255);
			MQFrame *frame3_1 = CreateHorizontalFrame(frame3);
			MQLabel *label = CreateLabel(frame3_1, L"U max");
			MQDoubleSpinBox *spin = CreateDoubleSpinBox(frame3_1);
			m_spin[2] = spin;
			spin->SetMin(0.0);
			spin->SetMax(1.0);
			spin->SetIncrement(0.01);
			spin->SetVisibleColumn(6);
			spin->AddChangingEvent(this, &NumericalManipulateDialog::change_spin);
			MQFrame *frame3_2 = CreateHorizontalFrame(frame3);
			MQButton *button_init = CreateButton(frame3_2, L"Init");
			button_init->AddClickEvent(this, &NumericalManipulateDialog::click_init);
			MQButton *button_min = CreateButton(frame3_2, L"min");
			button_min->AddClickEvent(this, &NumericalManipulateDialog::click_min);
			MQButton *button_mid = CreateButton(frame3_2, L"0.5");
			button_mid->AddClickEvent(this, &NumericalManipulateDialog::click_mid);
			MQButton *button_1 = CreateButton(frame3_2, L"1.0");
			button_1->AddClickEvent(this, &NumericalManipulateDialog::click_max);
			m_button_init[2] = button_init;
			m_button_min[2] = button_min;
			m_button_mid[2] = button_mid;
			m_button_max[2] = button_1;
		}
		{
			MQFrame *frame4 = CreateHorizontalFrame(mainframe);
			MQFrame *frame4_0 = CreateVerticalFrame(frame4);
			frame4_0->SetFillBeforeRate(0.4);
			frame4_0->SetFillAfterRate(0.4);
			frame4_0->SetBackColor(0, 0, 0, 255);
			MQFrame *frame4_1 = CreateHorizontalFrame(frame4_0);
			MQLabel *label = CreateLabel(frame4_1, L"V max");
			MQDoubleSpinBox *spin = CreateDoubleSpinBox(frame4_1);
			m_spin[3] = spin;
			spin->SetMin(0.0);
			spin->SetMax(1.0);
			spin->SetIncrement(0.01);
			spin->SetVisibleColumn(6);
			spin->AddChangingEvent(this, &NumericalManipulateDialog::change_spin);
			MQFrame *frame4_2 = CreateHorizontalFrame(frame4_0);
			MQButton *button_init = CreateButton(frame4_2, L"Init");
			button_init->AddClickEvent(this, &NumericalManipulateDialog::click_init);
			MQButton *button_min = CreateButton(frame4_2, L"min");
			button_min->AddClickEvent(this, &NumericalManipulateDialog::click_min);
			MQButton *button_mid = CreateButton(frame4_2, L"0.5");
			button_mid->AddClickEvent(this, &NumericalManipulateDialog::click_mid);
			MQButton *button_1 = CreateButton(frame4_2, L"1.0");
			button_1->AddClickEvent(this, &NumericalManipulateDialog::click_max);
			m_button_init[3] = button_init;
			m_button_min[3] = button_min;
			m_button_mid[3] = button_mid;
			m_button_max[3] = button_1;
		}
	}

	{
		MQFrame *frame6_0 = CreateHorizontalFrame(this);
		{
			MQFrame *frame6_1 = CreateHorizontalFrame(frame6_0);
			frame6_1->SetBackColor(0, 0, 0, 255);
			MQLabel *label5_1 = CreateLabel(frame6_1, L"Invert");
			m_button_horizontal = CreateButton(frame6_1, L"Horizontal");
			m_button_horizontal->AddClickEvent(this, &NumericalManipulateDialog::invert);
			m_button_vertical = CreateButton(frame6_1, L"Vertical");
			m_button_vertical->AddClickEvent(this, &NumericalManipulateDialog::invert);
		}
		{
			MQFrame *frame6_2 = CreateHorizontalFrame(frame6_0);
			frame6_2->SetBackColor(0, 0, 0, 255);
			MQLabel *label5_2 = CreateLabel(frame6_2, L"Turn");
			m_button_up = CreateButton(frame6_2, L"U");
			m_button_up->AddClickEvent(this, &NumericalManipulateDialog::turn);
			m_button_down = CreateButton(frame6_2, L"D");
			m_button_down->AddClickEvent(this, &NumericalManipulateDialog::turn);
			m_button_left = CreateButton(frame6_2, L"L");
			m_button_left->AddClickEvent(this, &NumericalManipulateDialog::turn);
			m_button_right = CreateButton(frame6_2, L"R");
			m_button_right->AddClickEvent(this, &NumericalManipulateDialog::turn);
		}
		{
			MQFrame *frame6_3 = CreateHorizontalFrame(frame6_0);
			frame6_3->SetBackColor(0, 0, 0, 255);
			m_button_rotate = CreateButton(frame6_3, L"Rotate");
			m_button_rotate->AddClickEvent(this, &NumericalManipulateDialog::rotate);
			m_spinangle = CreateDoubleSpinBox(frame6_3);
			m_spinangle->SetPosition(0.0);
			m_spinangle->SetMin(-180.0);
			m_spinangle->SetMax(180.0);
			m_spinangle->SetIncrement(5.0);
		}
	}
	{
		MQFrame *frame7 = CreateHorizontalFrame(this);
		{
			MQFrame *frame6_4 = CreateHorizontalFrame(frame7);
			frame6_4->SetBackColor(0, 0, 0, 255);
			m_button_movescale = CreateButton(frame6_4, L"Move");
			m_button_movescale->AddClickEvent(this, &NumericalManipulateDialog::move);
			m_spinmoveu = CreateDoubleSpinBox(frame6_4);
			m_spinmoveu->SetPosition(0.0);
			m_spinmoveu->SetMin(-1.0);
			m_spinmoveu->SetMax(1.0);
			m_spinmoveu->SetIncrement(0.1);
			m_spinmoveu->SetVisibleColumn(5);
			m_spinmovev = CreateDoubleSpinBox(frame6_4);
			m_spinmovev->SetPosition(0.0);
			m_spinmovev->SetMin(-1.0);
			m_spinmovev->SetMax(1.0);
			m_spinmovev->SetIncrement(0.1);
			m_spinmovev->SetVisibleColumn(5);
		}
		{
			MQFrame *frame6_5 = CreateHorizontalFrame(frame7);
			frame6_5->SetBackColor(0, 0, 0, 255);
			m_button_movescale = CreateButton(frame6_5, L"Scale");
			m_button_movescale->AddClickEvent(this, &NumericalManipulateDialog::scale);
			m_button_rotate->AddClickEvent(this, &NumericalManipulateDialog::rotate);
			m_spinscale = CreateDoubleSpinBox(frame6_5);
			m_spinscale->SetPosition(1.0);
			m_spinscale->SetMin(-10.0);
			m_spinscale->SetMax(10.0);
			m_spinscale->SetIncrement(0.1);
			m_spinscale->SetVisibleColumn(5);
		}
		MQButton *button_cancel = CreateButton(frame7, L"Cancel");
		button_cancel->SetFillBeforeRate(1.0);
		button_cancel->SetCancel(true);
		button_cancel->SetModalResult(DIALOG_CANCEL);
	}
	initializeUV();
}

void NumericalManipulateDialog::initializeUV()
{
	cmq::CMQDocument doc(m_doc);
	cmq::CMQObject obj = doc.getCurrentObject();

	m_pPlugin->getUVrange(obj, m_initumin, m_initumax, m_initvmin, m_initvmax);
	m_spin[0]->SetPosition(m_initvmin);
	m_spin[1]->SetPosition(m_initumin);
	m_spin[2]->SetPosition(m_initumax);
	m_spin[3]->SetPosition(m_initvmax);
	set_pivot();
}

void NumericalManipulateDialog::set_pivot(int index)
{
	if(index == -1 || index == 1 || index == 2) {
		double umin = m_spin[1]->GetPosition();
		double umax = m_spin[2]->GetPosition();
		m_spin[4]->SetPosition((umin + umax) / 2.0);
		set_movedelta(index);
	}
	if(index == -1 || index == 0 || index == 3) {
		double vmin = m_spin[0]->GetPosition();
		double vmax = m_spin[3]->GetPosition();
		m_spin[5]->SetPosition((vmin + vmax) / 2.0);
		set_movedelta(index);
	}
}

void NumericalManipulateDialog::set_movedelta(int index)
{
	if(index == -1 || index == 1 || index == 2 || index == 4) {
		float u = (float)m_spin[4]->GetPosition();
		float udelta = u - (m_initumin + m_initumax) / 2.0f;
		m_spinmoveu->SetPosition(udelta);
	}
	if(index == -1 || index == 0 || index == 3 || index == 5) {
		float v = (float)m_spin[5]->GetPosition();
		float vdelta = v - (m_initvmin + m_initvmax) / 2.0f;
		m_spinmovev->SetPosition(vdelta);
	}
}

BOOL NumericalManipulateDialog::change_spin(MQWidgetBase *sender, MQDocument doc)
{
	set_pivot();
	return FALSE;
}

BOOL NumericalManipulateDialog::change_pivot(MQWidgetBase *sender, MQDocument doc)
{
	set_movedelta((sender == m_spin[4]) ? 4 : 5);
	return FALSE;
}

BOOL NumericalManipulateDialog::click_init(MQWidgetBase *sender, MQDocument doc)
{
	MQButton *button = dynamic_cast<MQButton*>(sender);
	if(!button) { return FALSE; }
	if(button == m_button_init[0]) {
		m_spin[0]->SetPosition(m_initvmin);
		set_pivot(0);
	}
	else if(button == m_button_init[1]) {
		m_spin[1]->SetPosition(m_initumin);
		set_pivot(1);
	}
	else if(button == m_button_init[2]) {
		m_spin[2]->SetPosition(m_initumax);
		set_pivot(2);
	}
	else if(button == m_button_init[3]) {
		m_spin[3]->SetPosition(m_initvmax);
		set_pivot(3);
	}
	else if(button == m_button_init[4]) {
		m_spin[4]->SetPosition((m_initumax + m_initumin) / 2.0);
		set_movedelta(4);
	}
	else if(button == m_button_init[5]) {
		m_spin[5]->SetPosition((m_initvmax + m_initvmin) / 2.0);
		set_movedelta(5);
	}
	return FALSE;
}

BOOL NumericalManipulateDialog::click_min(MQWidgetBase *sender, MQDocument doc)
{
	MQButton *button = dynamic_cast<MQButton*>(sender);
	for(int i = 0; i < 6; i++) {
		if(button == m_button_min[i]) {
			if(i <= 1) {
				m_spin[i]->SetPosition(0.0);
				set_pivot(i);
			}
			else if(i < 4) {
				m_spin[i]->SetPosition(m_spin[3 - i]->GetPosition());
				set_pivot(i);
			}
			else {
				m_spin[i]->SetPosition(m_spin[5 - i]->GetPosition());
				set_movedelta(i);
			}
			break;
		}
	}
	return FALSE;
}

BOOL NumericalManipulateDialog::click_mid(MQWidgetBase *sender, MQDocument doc)
{
	MQButton *button = dynamic_cast<MQButton*>(sender);
	for(int i = 0; i < 6; i++) {
		if(button == m_button_mid[i]) {
			m_spin[i]->SetPosition(0.5);
			set_pivot(i);
			set_movedelta(i);
		}
	}
	return FALSE;
}

BOOL NumericalManipulateDialog::click_max(MQWidgetBase *sender, MQDocument doc)
{
	MQButton *button = dynamic_cast<MQButton*>(sender);
	for(int i = 0; i < 6; i++) {
		if(button == m_button_max[i]) {
			if(i <= 1) {
				m_spin[i]->SetPosition(m_spin[3 - i]->GetPosition());
				set_pivot(i);
			}
			else if(i < 4) {
				m_spin[i]->SetPosition(1.0);
				set_pivot(i);
			}
			else {
				m_spin[i]->SetPosition(m_spin[i - 2]->GetPosition());
				set_movedelta(i);
			}
			break;
		}
	}
	return FALSE;
}

BOOL NumericalManipulateDialog::invert(MQWidgetBase *sender, MQDocument doc)
{
	UVToolsX::invert_direction_t direction = UVToolsX::INVERT_NO;
	if(sender == m_button_horizontal) {
		direction = UVToolsX::INVERT_HORIZONTAL;
	}
	else {
		direction = UVToolsX::INVERT_VERTICAL;
	}
	float vmin = (float)m_spin[0]->GetPosition();
	float umin = (float)m_spin[1]->GetPosition();
	float umax = (float)m_spin[2]->GetPosition();
	float vmax = (float)m_spin[3]->GetPosition();
	return m_pPlugin->invertUV(doc, umin, umax, vmin, vmax, direction);
}

BOOL NumericalManipulateDialog::turn(MQWidgetBase *sender, MQDocument doc)
{
	UVToolsX::turn_direction_t direction = UVToolsX::TURN_NO;
	if(sender == m_button_up) { direction = UVToolsX::TURN_UP; }
	else if(sender == m_button_down) { direction = UVToolsX::TURN_DOWN; }
	else if(sender == m_button_left) { direction = UVToolsX::TURN_LEFT; }
	else if(sender == m_button_right) { direction = UVToolsX::TURN_RIGHT; }
	float vmin = (float)m_spin[0]->GetPosition();
	float umin = (float)m_spin[1]->GetPosition();
	float umax = (float)m_spin[2]->GetPosition();
	float vmax = (float)m_spin[3]->GetPosition();
	return m_pPlugin->turnUV(doc, umin, umax, vmin, vmax, direction);
}

BOOL NumericalManipulateDialog::rotate(MQWidgetBase *sender, MQDocument doc)
{
	float u = (float)m_spin[4]->GetPosition();
	float v = (float)m_spin[5]->GetPosition();
	MQCoordinate pivot(u, v);
	float angle = (float)m_spinangle->GetPosition();
	return m_pPlugin->rotate(doc, pivot, angle);
}

BOOL NumericalManipulateDialog::move(MQWidgetBase *sender, MQDocument doc)
{
	float udelta = (float)m_spinmoveu->GetPosition();
	float vdelta = (float)m_spinmovev->GetPosition();
	MQCoordinate delta(udelta, vdelta);
	return m_pPlugin->move(doc, delta);
}

BOOL NumericalManipulateDialog::scale(MQWidgetBase *sender, MQDocument doc)
{
	float u = (float)m_spin[4]->GetPosition();
	float v = (float)m_spin[5]->GetPosition();
	MQCoordinate pivot(u, v);
	float rate = (float)m_spinscale->GetPosition();
	return m_pPlugin->scale(doc, pivot, rate);
}

//--------------------------------------------------

TextureOutputDialog::TextureOutputDialog(MQWindowBase& parent, UVToolsX *plugin) : MQDialog(parent), m_pPlugin(plugin)
{
	setlocale(LC_CTYPE, "");
	SetTitle(L"FileOuput");
	this->SetOutSpace(0.0);
	this->SetInSpace(0.0);

	MQFrame *mainFrame = CreateVerticalFrame(this);

	{
		MQFrame *frame1 = CreateHorizontalFrame(mainFrame);
		frame1->SetMultiColumn(true);
		frame1->SetMatrixColumn(2);

		m_spinwidth = CreateSpinBox(frame1);
		m_spinwidth->SetMin(1);
		m_spinwidth->SetMax(65535);
		m_spinwidth->SetVisibleColumn(5);
		m_spinwidth->SetIncrement(4);
		m_spinwidth->AddChangedEvent(this, &TextureOutputDialog::texsizechanged);

		m_spinheight = CreateSpinBox(frame1);
		m_spinheight->SetMin(1);
		m_spinheight->SetMax(65535);
		m_spinheight->SetVisibleColumn(5);
		m_spinheight->SetIncrement(4);
		m_spinheight->AddChangedEvent(this, &TextureOutputDialog::texsizechanged);
	}
	{
		MQFrame *frame2 = CreateHorizontalFrame(mainFrame);
		frame2->SetMultiColumn(true);
		frame2->SetMatrixColumn(2);

		m_checkwire = CreateCheckBox(frame2, L"wireframe");
		m_checkwire->SetChecked(!!(m_pPlugin->m_texturemode & UVToolsX::TEX_WIRE));
#ifdef EX
		m_checkvertexcol = CreateCheckBox(frame2, L"vertex color");
		m_checkvertexcol->SetChecked(!!(m_pPlugin->m_texturemode & UVToolsX::TEX_VERT));
#endif
		m_checkmatcol = CreateCheckBox(frame2, L"material color");
		m_checkmatcol->SetChecked(!!(m_pPlugin->m_texturemode & UVToolsX::TEX_MAT));
		m_checkobjcol = CreateCheckBox(frame2, L"object color");
		m_checkobjcol->SetChecked(!!(m_pPlugin->m_texturemode & UVToolsX::TEX_OBJ));
		m_checkshader = CreateCheckBox(frame2, L"shader");
		m_checkshader->SetChecked(!!(m_pPlugin->m_texturemode & UVToolsX::TEX_SHADER));
	}
	{
		MQFrame *frame3 = CreateHorizontalFrame(mainFrame);
		m_checkmemory = CreateCheckBox(frame3, L"memory");
		{
			MQFrame *frame2_2 = CreateHorizontalFrame(frame3);
			for(int i = 0; i < m_pPlugin->STORESIZE; i++) {
				wchar_t no[2];
				wsprintf(no, L"%d", i + 1);
				m_checkmemorystore.push_back(CreateCheckBox(frame2_2, no));
			}
		}
	}
	{
		MQFrame *frame4 = CreateHorizontalFrame(mainFrame);
		m_currentobj = CreateCheckBox(frame4, L"current object");
		m_currentobj->SetChecked(!!(m_pPlugin->m_texturemode & UVToolsX::TEX_CUROBJ));
		m_currentmat = CreateCheckBox(frame4, L"current material");
		m_currentmat->SetChecked(!!(m_pPlugin->m_texturemode & UVToolsX::TEX_CURMAT));
	}
	{
		MQFrame *frame5 = CreateHorizontalFrame(mainFrame);
		MQButton *buttonCilpboardColor = CreateButton(frame5, L"Color to Clipboard");
		buttonCilpboardColor->AddClickEvent(this, &TextureOutputDialog::clipcolor);

		MQButton *buttonCilpboardAlpha = CreateButton(frame5, L"Alpha to Clipboard");
		buttonCilpboardAlpha->AddClickEvent(this, &TextureOutputDialog::clipalpha);
	}
	{
		MQFrame *frame3 = CreateHorizontalFrame(mainFrame);
		m_editfilename = CreateEdit(frame3);
		m_editfilename->SetHorzLayout(LAYOUT_FILL);
		MQButton *buttonfile = CreateButton(frame3, L"refer");
		buttonfile->AddClickEvent(this, &TextureOutputDialog::savedialog);

		MQFrame *frame6 = CreateHorizontalFrame(mainFrame);
		MQButton *buttonOK = CreateButton(frame6, L"OK");
		buttonOK->SetModalResult(DIALOG_OK);
		buttonOK->SetDefault(true);
		MQButton *buttonCancel = CreateButton(frame6, L"Cancel");
		buttonCancel->SetCancel(true);
		buttonCancel->SetModalResult(DIALOG_CANCEL);
	}
}

BOOL TextureOutputDialog::texsizechanged(MQWidgetBase *sender, MQDocument doc)
{
	MQSpinBox *spin = dynamic_cast<MQSpinBox*>(sender);
	if(!spin) {
		return FALSE;
	}
	int pos = spin->GetPosition();
	pos -= pos % 4;
	spin->SetPosition(pos);
	return FALSE;
}


BOOL TextureOutputDialog::clipcolor(MQWidgetBase *sender, MQDocument doc)
{
	DWORD mode = (m_checkwire->GetChecked() ? UVToolsX::TEX_WIRE : 0) |
		(m_checkvertexcol && m_checkvertexcol->GetChecked() ? UVToolsX::TEX_VERT : 0) |
		(m_checkmatcol->GetChecked() ? UVToolsX::TEX_MAT : 0) |
		(m_checkobjcol->GetChecked() ? UVToolsX::TEX_OBJ : 0) |
		(m_checkshader->GetChecked() ? UVToolsX::TEX_SHADER : 0) |
		(m_currentobj->GetChecked() ? UVToolsX::TEX_CUROBJ : 0) |
		(m_currentmat->GetChecked() ? UVToolsX::TEX_CURMAT : 0);
	mode |= UVToolsX::CLIPCOLOR;
	DWORD memory = 0;
	if(m_checkmemory->GetChecked()) {
		for(int i = 0; i < UVToolsX::STORESIZE; i++) {
			if(m_checkmemorystore[i]->GetChecked()) {
				memory |= (1 << i);
			}
		}
	}
	m_pPlugin->writefile(doc, m_spinwidth->GetPosition(), m_spinheight->GetPosition(), L"", mode, memory);
	return FALSE;
}

BOOL TextureOutputDialog::clipalpha(MQWidgetBase *sender, MQDocument doc)
{
	DWORD mode = (m_checkwire->GetChecked() ? UVToolsX::TEX_WIRE : 0) |
		(m_checkvertexcol && m_checkvertexcol->GetChecked() ? UVToolsX::TEX_VERT : 0) |
		(m_checkmatcol->GetChecked() ? UVToolsX::TEX_MAT : 0) |
		(m_checkobjcol->GetChecked() ? UVToolsX::TEX_OBJ : 0) |
		(m_checkshader->GetChecked() ? UVToolsX::TEX_SHADER : 0) |
		(m_currentobj->GetChecked() ? UVToolsX::TEX_CUROBJ : 0) |
		(m_currentmat->GetChecked() ? UVToolsX::TEX_CURMAT : 0);
	mode |= UVToolsX::CLIPALPHA;
	DWORD memory = 0;
	if(m_checkmemory->GetChecked()) {
		for(int i = 0; i < UVToolsX::STORESIZE; i++) {
			if(m_checkmemorystore[i]->GetChecked()) {
				memory |= (1 << i);
			}
		}
	}
	m_pPlugin->writefile(doc, m_spinwidth->GetPosition(), m_spinheight->GetPosition(), L"", mode, memory);
	return FALSE;
}

BOOL TextureOutputDialog::savedialog(MQWidgetBase *sender, MQDocument doc)
{
	MQSaveFileDialog dlg(*this);
	dlg.SetTitle(L"Texture filename");
	dlg.AddFilter(L"Bitmap file (*.bmp)|*.bmp");
	dlg.AddFilter(L"Tegra file (*.tga)|*.tga");
	dlg.AddFilter(L"Jpeg file (*.jpg)|*.jpg;*.jpeg");
	dlg.AddFilter(L"PNG file (*.png)|*.png");
	dlg.SetFilterIndex(0);
	dlg.SetFileName(m_editfilename->GetText());
	if(!dlg.Execute()) {
		return FALSE;
	}
	m_editfilename->SetText(dlg.GetFileName());
	return TRUE;
}
