#include "MQWidget.h"
#include "CMQObject.h"

#ifndef _UIWindowDefine
#define _UIWindowDefine

class UVToolsX;

class UVToolsXWindow : public MQWindow
{
	friend class NumericalManipulateDialog;
public:
	UVToolsXWindow(UVToolsX *plugin, MQWindowBase& parent);
	~UVToolsXWindow();

	void resize();
	BOOL resize(MQWidgetBase *sender, MQDocument doc) {resize(); return FALSE;}

	void clearbutton();
	void setbutton(int index, bool flag = true);
	UVToolsX *m_pPlugin;
private:

	BOOL sel_floating(MQWidgetBase *sender, MQDocument doc);
	BOOL sel_group(MQWidgetBase *sender, MQDocument doc);
	BOOL sel_samevert(MQWidgetBase *sender, MQDocument doc);

	BOOL selUV2OBJvert(MQWidgetBase *sender, MQDocument doc);
	BOOL selUV2OBJline(MQWidgetBase *sender, MQDocument doc);
	BOOL selUV2OBJface(MQWidgetBase *sender, MQDocument doc);
	BOOL selOBJ2UVvert(MQWidgetBase *sender, MQDocument doc);
	BOOL selOBJ2UVline(MQWidgetBase *sender, MQDocument doc);
	BOOL selOBJ2UVface(MQWidgetBase *sender, MQDocument doc);

	BOOL grab_vert(MQWidgetBase *sender, MQDocument doc);
	BOOL grab_face(MQWidgetBase *sender, MQDocument doc);
	BOOL grab_regular(MQWidgetBase *sender, MQDocument doc);

	BOOL invert(MQWidgetBase *sender, MQDocument doc);
	BOOL turn(MQWidgetBase *sender, MQDocument doc);

	BOOL sew(MQWidgetBase *sender, MQDocument doc);
	BOOL gather(MQWidgetBase *sender, MQDocument doc);

	BOOL regularize(MQWidgetBase *sender, MQDocument doc);
	BOOL symmetrize(MQWidgetBase *sender, MQDocument doc);
	BOOL numerical(MQWidgetBase *sender, MQDocument doc);

	BOOL set(MQWidgetBase *sender, MQDocument doc);
	BOOL restore(MQWidgetBase *sender, MQDocument doc);
	BOOL lock(MQWidgetBase *sender, MQDocument doc, MQWidgetMouseParam& param);

	BOOL textureoutput(MQWidgetBase *sender, MQDocument doc);

	BOOL OnHide(MQWidgetBase *sender, MQDocument doc);
	BOOL keyevent(MQWidgetBase *sender, MQDocument doc, MQWidgetKeyParam& param);

	MQButton *button_grab_vert;
	MQButton *button_grab_face;
	MQButton *button_grab_group;
	MQButton *button_grab_regular;

	MQButton *m_button_invert_h;
	MQButton *m_button_invert_v;
	MQButton *m_button_turn_up;
	MQButton *m_button_turn_down;
	MQButton *m_button_turn_left;
	MQButton *m_button_turn_right;

	MQCheckBox *m_check_symmetrize_mode;
	MQDoubleSpinBox *m_spin_symmetirize_delta;

	MQCheckBox *m_check_selectedfacemode;

	std::vector<MQButton*> m_button_store[2]; //[UVToolsX::STORESIZE];
	std::vector<MQButton*> m_button_lock; //[UVToolsX::STORESIZE];

	cmq::CMQObject obj_;
};

class NumericalManipulateDialog : public MQDialog
{
public:
	NumericalManipulateDialog(MQWindowBase& parent, UVToolsX *plugin, MQDocument doc);
	~NumericalManipulateDialog() {}
private:
	UVToolsX *m_pPlugin;
	MQDocument m_doc;
	float m_initumin, m_initumax, m_initvmin, m_initvmax;

	MQDoubleSpinBox *m_spin[6];

	MQButton *m_button_init[6];
	MQButton *m_button_min[6];
	MQButton *m_button_max[6];
	MQButton *m_button_mid[6];

	MQButton *m_button_horizontal;
	MQButton *m_button_vertical;
	MQButton *m_button_up;
	MQButton *m_button_down;
	MQButton *m_button_left;
	MQButton *m_button_right;

	MQButton *m_button_rotate;
	MQDoubleSpinBox *m_spinangle;
	MQDoubleSpinBox *m_spinmoveu;
	MQDoubleSpinBox *m_spinmovev;
	MQDoubleSpinBox *m_spinscale;

	MQButton *m_button_movescale;

	void initializeUV();
	void set_pivot(int type = -1);
	void set_movedelta(int index);

	BOOL click_init(MQWidgetBase *sender, MQDocument doc);
	BOOL click_min(MQWidgetBase *sender, MQDocument doc);
	BOOL click_mid(MQWidgetBase *sender, MQDocument doc);
	BOOL click_max(MQWidgetBase *sender, MQDocument doc);
	BOOL change_spin(MQWidgetBase *sender, MQDocument doc);
	BOOL change_pivot(MQWidgetBase *sender, MQDocument doc);

	BOOL turn(MQWidgetBase *sender, MQDocument doc);
	BOOL invert(MQWidgetBase *sender, MQDocument doc);
	BOOL rotate(MQWidgetBase *sender, MQDocument doc);
	BOOL move(MQWidgetBase *sender, MQDocument doc);
	BOOL scale(MQWidgetBase *sender, MQDocument doc);
};

class TextureOutputDialog : public MQDialog
{
public:
	TextureOutputDialog(MQWindowBase& parent, UVToolsX *plugin);
	~TextureOutputDialog() {}

	UVToolsX *m_pPlugin;
private:
	MQSpinBox *m_spinwidth;
	MQSpinBox *m_spinheight;
	MQCheckBox *m_checkwire;
	MQCheckBox *m_checkvertexcol;
	MQCheckBox *m_checkmatcol;
	MQCheckBox *m_checkobjcol;
	MQCheckBox *m_checkshader;
	MQCheckBox *m_currentobj;
	MQCheckBox *m_currentmat;

	MQCheckBox *m_checkmemory;
	std::vector<MQCheckBox*> m_checkmemorystore;

	MQEdit *m_editfilename;

	BOOL clipcolor(MQWidgetBase *sender, MQDocument doc);
	BOOL clipalpha(MQWidgetBase *sender, MQDocument doc);

	BOOL texsizechanged(MQWidgetBase *sender, MQDocument doc);

	BOOL savedialog(MQWidgetBase *sender, MQDocument doc);

	friend class UVToolsXWindow;
};

#endif
